import logging
import re
from datetime import date
from functools import partial

import orjson
from nicegui import ElementFilter, app, events, run, ui
from nicegui.binding import BindableProperty
from pydantic import AnyUrl, ValidationError
from xplan_tools.model import model_factory
from xplan_tools.model.base import BaseFeature
from xplan_tools.model.orm import Feature
from xplan_tools.util.style import add_style_properties_to_feature

from xmas_app.components.association import AssociationList
from xmas_app.components.documents import DocumentSelect
from xmas_app.components.label import LabelWithDescription
from xmas_app.settings import get_settings
from xmas_app.util.codelist import get_codelist_options

logger = logging.getLogger("xmas_app")


class ModelForm:
    """A dynamic form builder for XPlan features using NiceGUI.

    This class provides methods to render and manage forms for editing and previewing
    XPlan/XTrasse features, including nested data types and associations.
    """

    editable: bool = BindableProperty()

    def __init__(
        self,
        appschema: str,
        version: str,
        parent: ui.element | None = None,
        header: ui.header | None = None,
    ):
        """Initialize the ModelForm.

        Args:
            appschema (str): The application schema name.
            version (str): The schema version.
            parent (ui.element | None): The parent UI element for rendering.
            header (ui.header | None): Optional header UI element.
        """
        self.appschema = appschema
        self.version = version
        self.featuretype = None
        self.sub_models = {}
        self.parent = parent
        self.header = header
        self.rendered = False
        self.editable = False
        self.validation_errors = []
        self.validation_notification = None
        ui.on("showValidationDetails", self._show_validation_details)

    async def _show_validation_details(self, _: events.GenericEventArguments):
        def _go_to_property_row(key: str):
            validation_dialog.close()
            for element in ElementFilter(marker=["property_row", key]).not_within(
                marker="sub_form"
            ):
                ui.navigate.to(element)
                break

        if self.parent:
            with self.parent, ui.dialog() as validation_dialog, ui.card():
                ui.label("Validierungsfehler").classes("text-bold")
                with ui.list():
                    for error in self.validation_errors:
                        name = error["loc"][0]
                        with ui.item(
                            on_click=lambda key=name: _go_to_property_row(key)
                        ).tooltip(f"Klicken um zum Attribut {name!r} zu springen"):
                            with ui.item_section().props("avatar"):
                                ui.icon("error", color="negative")
                            with ui.item_section():
                                ui.item_label(name)
                                ui.item_label(error["msg"]).props("caption")
            await validation_dialog
            validation_dialog.delete()

    async def _filter_properties(self):
        """Filter property rows in the UI based on the selected filter option."""
        self.row_filter.set_value("")
        match self.radio_filter.value:
            case "all":
                ElementFilter(marker="property_row").classes(remove="hidden")
            case "mandatory":
                ElementFilter(marker="property_row").classes(remove="hidden")
                ElementFilter(marker="nullable").classes(add="hidden")
            case "set":
                ElementFilter(marker="property_row").classes(remove="hidden")
                for elem in (
                    ElementFilter(kind=ui.label)
                    .within(marker="property_row")
                    .not_within(marker="sub_form")
                ):
                    if not getattr(self.feature, elem.text, None):
                        ElementFilter(marker=elem.text).not_within(
                            marker="sub_form"
                        ).classes(add="hidden")

    @staticmethod
    def validate_uri(value: str | None):
        if value is None:
            return True
        try:
            _ = AnyUrl(value)
        except ValidationError:
            return False
        else:
            return True

    @staticmethod
    def validate_date(value: str | None):
        if value is None:
            return True
        try:
            date.fromisoformat(value)
        except ValueError:
            return False
        else:
            return True

    async def _preview_association(self, assoc_id: str):
        """Show a dialog previewing the associated object.

        Args:
            assoc_id (str): The ID of the associated object to preview.
        """
        with self.parent:
            with (
                ui.dialog() as dialog,
                ui.card().style("width: auto; max-width: none") as card,
            ):
                try:
                    linked_feature = get_settings().repo.get(assoc_id)
                    await ModelForm(
                        self.appschema,
                        self.version,
                        card,
                    ).render_form(
                        feature_type=linked_feature.get_name(),
                        feature=linked_feature.model_dump(
                            mode="json", context={"datatype": True, "file_uri": True}
                        ),
                        main_form=True,
                        preview=True,
                    )
                except ValidationError:
                    ui.notify(
                        "Fehler beim Anzeigen der Daten",
                        type="negative",
                        position="top",
                    )
        await dialog
        dialog.delete()

    def _propagate_data(self):
        if not self.editable:
            return
        if model_instance := self.model_instance:
            feature_data = {
                "properties": model_instance.model_dump(
                    mode="json",
                    exclude_unset=True,
                    exclude={"id", self.model.get_geom_field()},
                    context={"datatype": True},
                ),
                "featuretype": self.model.get_name(),
                "id": model_instance.id,
            }
        else:
            feature_data = {
                "properties": None,
                "featuretype": None,
                "id": None,
            }
            # self.radio_filter.set_value("mandatory")
        ui.run_javascript(f"""new QWebChannel(qt.webChannelTransport, function (channel) {{
                                    channel.objects.handler.receive_feature({orjson.dumps(feature_data).decode()})
                            }});""")

    def _get_bindable_object(self, model: type[BaseFeature]) -> type:
        """Construct a class with bindable properties from a BaseFeature.

        Takes a BaseFeature class as input and returns a new class with the BaseFeature's attributes as BindableProperties.
        The new bindable class can be instantiated with keyword arguments corresponding to the attributes,
        e.g. with an unpacked dict generated by the model_dump() method.

        Nested models will be assigned to corresponding bindable classes recursively.

        Args:
            model (type[BaseFeature]): The Pydantic BaseFeature to build the bindable class from.

        Returns:
            type: The bindable class with properties corresponding to the model fields.
        """

        def constructor(bindable_object, **kwargs) -> None:
            """The init method of the bindable class.

            Args:
                bindable_object (object): The class instance, i.e. self.
                kwargs: The feature's attributes as keyword arguments.
            """
            for key in model.model_fields.keys():
                value = kwargs.get(key, None)
                prop_info = model.get_property_info(key)
                if prop_info["stereotype"] == "DataType" and value:
                    if isinstance(prop_info["typename"], str):
                        sub_obj = self._get_bindable_object(
                            model_factory(
                                prop_info["typename"], self.version, self.appschema
                            )
                        )
                        value = (
                            [sub_obj(**item) for item in value]
                            if prop_info["list"]
                            else sub_obj(**value)
                        )
                    else:
                        value = (
                            [
                                self._get_bindable_object(
                                    model_factory(
                                        item["datatype"], self.version, self.appschema
                                    )
                                )(**item)
                                for item in value
                            ]
                            if prop_info["list"]
                            else self._get_bindable_object(
                                model_factory(
                                    value["datatype"], self.version, self.appschema
                                )
                            )(**value)
                        )
                setattr(bindable_object, key, value)

        attributes = {"__init__": constructor}
        for field in model.model_fields.keys():
            attributes[field] = BindableProperty(
                on_change=lambda *_: self._propagate_data()
            )
        return type(model.get_name(), (), attributes)

    @property
    def model_instance(self) -> BaseFeature | None:
        """Get the validated model instance from the current feature.

        Returns:
            BaseFeature | None: The validated model instance, or None if validation fails.
        """
        feature = self.feature
        vars(feature).pop("sub_models", None)
        if (geom := vars(self.feature).pop("geometry", None)) and (
            geom_field := self.model.get_geom_field()
        ):
            setattr(feature, geom_field, geom)
        try:
            model_instance = self.model.model_validate(feature)
            if self.validation_notification:
                self.validation_notification.dismiss()
            return model_instance
        except ValidationError as e:
            logger.debug(
                "validation error for feature %r with id %r:\n    %s",
                self.featuretype,
                self.feature.id,
                "\n    ".join(str(error) for error in e.errors()),
            )
            self.validation_errors = e.errors()
            if not self.validation_notification:
                self.validation_notification = ui.notification(
                    "Validierungsfehler",
                    type="negative",
                    position="bottom",
                    timeout=None,
                    on_dismiss=lambda: setattr(self, "validation_notification", None),
                    actions=[
                        {
                            "icon": "description",
                            ":handler": "() => {emitEvent('showValidationDetails')}",
                            "noDismiss": True,
                        },
                        {
                            "icon": "close",
                        },
                    ],
                )

    async def _set_stylesheetId_and_schriftinhalt(self) -> None:
        """Set the stylesheetId and schriftinhalt attributes based on the referenced feature."""
        if not self.editable:
            return
        ref_id = getattr(self.feature, "dientZurDarstellungVon", None)
        if not ref_id or not getattr(self.feature, "art", None):
            return
        obj = self.model_instance
        if not obj:
            return
        new_obj = add_style_properties_to_feature(
            obj,
            get_settings().repo.get(ref_id[0]),
            always_populate_schriftinhalt=get_settings().generate_schriftinhalt,
        )
        stylesheet_id = getattr(new_obj, "stylesheetId", None)
        setattr(
            self.feature, "stylesheetId", str(stylesheet_id) if stylesheet_id else None
        )
        if schriftinhalt := getattr(new_obj, "schriftinhalt", None):
            setattr(self.feature, "schriftinhalt", schriftinhalt)

    def _get_association_targets(
        self,
        property: str | list,
        typenames: str | list,
        excluded_ids: list[str] | str = [],
        plan_id: str | None = None,
    ):
        """Get possible association targets for a property.

        Args:
            property (str | list): The property name or list of names.
            typenames (str | list): The type name(s) to filter by.
            excluded_ids (list[str] | str, optional): IDs to exclude. Defaults to [].
            plan_id (str | None, optional): The ID of the associated plan. Defaults to None.

        Returns:
            dict: Mapping of feature IDs to feature type names.
        """
        if plan_id and isinstance(property, str) and property.startswith("gehoertZu"):
            plan = get_settings().repo.get(plan_id)
            if property.endswith("Bereich"):
                bereiche = [
                    get_settings().repo.get(bereich) for bereich in plan.bereich
                ]
                return {
                    bereich.id: f"{bereich.get_name()} {bereich.nummer}"
                    for bereich in bereiche
                }
            else:
                return {plan_id: f"{plan.get_name()} {plan.name}"}
        with (repo := get_settings().repo).Session() as session:
            stmt = (
                repo.get_feature_graph(self.feature.id)
                .where(
                    Feature.featuretype.in_(
                        [typenames] if isinstance(typenames, str) else typenames
                    )
                )
                .where(
                    Feature.id.not_in(
                        [excluded_ids]
                        if isinstance(excluded_ids, str)
                        else excluded_ids
                    )
                )
                .where(Feature.id != self.feature.id)
            )
            result = session.execute(stmt)

            targets = {
                str(feature.id): f"{feature.featuretype} {str(feature.id)[:8]}"
                for feature in result.unique().scalars().all()
            }

        return targets

    def _is_geometry_required(self, typename: str) -> bool:
        """Check if a geometry is required for the given type.

        Args:
            typename (str): The type name.

        Returns:
            bool: True if geometry is required, False otherwise.
        """
        model = model_factory(typename, self.version, self.appschema)
        if geom := model.get_geom_field():
            return not model.get_property_info(geom)["nullable"]
        return False

    def _get_art_options(self) -> None:
        """Populate the 'art' select input with options derived from referenced feature attributes."""

        def derive_art_from_attributes():
            """Derive art options from the referenced feature's attributes.

            Returns:
                dict: Mapping of attribute paths to display values.
            """

            def set_option_value(path, value):
                art[path] = f"{path} ({str(value)})"

            art = {}
            for name, value in ref_feature.model_dump(
                mode="json",
                exclude_none=True,
                exclude={"id", ref_feature.get_geom_field()},
            ).items():
                # TODO refactor recursive parsing
                prop_info = ref_feature.get_property_info(name)
                if prop_info["stereotype"] == "Association":
                    continue
                elif prop_info["list"]:
                    for index, item in enumerate(value, 0):
                        path = f"{name}[{index + 1}]"
                        if prop_info["stereotype"] == "BasicType":
                            set_option_value(path, item)
                        elif prop_info["stereotype"] == "Enumeration":
                            set_option_value(path, prop_info["enum_info"][item]["name"])
                        elif prop_info["stereotype"] == "DataType":
                            sub_model = getattr(ref_feature, name)[index]
                            for sub_name, sub_value in item.items():
                                sub_prop_info = sub_model.get_property_info(sub_name)
                                sub_path = "/".join(
                                    [path, sub_model.get_name(), sub_name]
                                )
                                set_option_value(
                                    sub_path,
                                    sub_prop_info["enum_info"][sub_value]["name"]
                                    if sub_prop_info["stereotype"] == "Enumeration"
                                    else sub_value,
                                )
                elif prop_info["stereotype"] == "DataType":
                    sub_model = model_factory(
                        prop_info["typename"],
                        self.version,
                        self.appschema,
                    )
                    for sub_name, sub_value in value.items():
                        sub_prop_info = sub_model.get_property_info(sub_name)
                        path = "/".join([name, prop_info["typename"], sub_name])
                        set_option_value(
                            path,
                            sub_prop_info["enum_info"][sub_value]["name"]
                            if sub_prop_info["stereotype"] == "Enumeration"
                            else sub_value,
                        )
                else:
                    set_option_value(
                        name,
                        prop_info["enum_info"][value]["name"]
                        if prop_info["stereotype"] == "Enumeration"
                        else value,
                    )
            return art

        if ref := getattr(self.feature, "dientZurDarstellungVon", None):
            ref_feature = get_settings().repo.get(ref[0])
            self.feature.gehoertZuBereich = str(ref_feature.gehoertZuBereich)
            art_options = derive_art_from_attributes()
            self.art_input.set_options(self.art_input.options | art_options)

    # @ui.refreshable_method
    async def _build_sub_form(
        self, key: str, bind_obj: type, prop_info: dict, preview: bool = False
    ) -> None:
        """Construct a nested form for data types.

        Args:
            key (str): The name of the attribute where the data type is the value.
            bind_obj (type): The bindable parent class the sub-form is bound to.
            prop_info (dict): The property info object for the given attribute.
            preview (bool, optional): Whether the form should be rendered in preview mode. Defaults to False.
        """

        async def add_form(type_name: str | None = None):
            """Add a new sub-form instance for the data type.

            Args:
                type_name (str | None): The type name to instantiate, if it's a union type, otherwise the typename is taken from the property info.
            """
            sub_obj = self._get_bindable_object(
                model_factory(
                    type_name or prop_info["typename"], self.version, self.appschema
                )
            )()
            if not getattr(bind_obj, key, None):
                setattr(bind_obj, key, [sub_obj] if prop_info["list"] else sub_obj)
            elif prop_info["list"]:
                getattr(bind_obj, key).append(sub_obj)
            else:
                setattr(bind_obj, key, sub_obj)

            item = await build_item(
                None,
                sub_obj,
                getattr(bind_obj, key).index(sub_obj) if prop_info["list"] else None,
            )
            item.move(item_list)
            # self._build_sub_form.refresh()

        def delete_form(element: ui.element, item: type):
            """Delete a sub-form instance.

            Args:
                index (int | None): The index to delete, if the property is a list.
            """
            if isinstance(getattr(bind_obj, key), list):
                getattr(bind_obj, key).remove(item)
                if not len(getattr(bind_obj, key)):
                    setattr(bind_obj, key, None)
            else:
                setattr(bind_obj, key, None)
            element.delete()
            # self._build_sub_form.refresh()

        async def build_item(
            sub_model: BaseFeature | None,
            item: type,
            index: int | None = None,
        ) -> ui.item:
            """Build a UI item for a sub-form instance.

            This corresponds to an instance of the data type.

            Args:
                sub_model (BaseFeature | None): The sub-model class.
                item (type): The bindable object instance.
                index (int | None): The index in the list, if applicable.
            """
            if not sub_model:
                sub_model = model_factory(
                    item.__class__.__name__,
                    self.version,
                    self.appschema,
                )

            with ui.item().props("dense") as list_item:
                with ui.item_section():
                    if isinstance(prop_info["typename"], list):
                        with ui.item_label():
                            LabelWithDescription(
                                sub_model.get_name(), sub_model.__doc__
                            ).classes("text-bold")
                    await self._build_model_form(
                        sub_model,
                        getattr(bind_obj, key)[index]
                        if prop_info["list"]
                        else getattr(bind_obj, key),
                        preview=preview,
                    )
                if not preview:
                    with ui.item_section().props("side"):
                        delete_button = ui.button(
                            icon="delete",
                            on_click=partial(delete_form, list_item, item),
                        ).props("flat round")
                        if not self.editable or (
                            not prop_info["nullable"]
                            and prop_info["list"]
                            and index == 0
                        ):
                            delete_button.disable()
                        else:
                            delete_button.bind_enabled_from(
                                bind_obj,
                                key,
                                backward=lambda x: (
                                    len(x) > 1
                                    if not prop_info["nullable"] and isinstance(x, list)
                                    else prop_info["nullable"]
                                ),
                            )

            return list_item

        if not prop_info["nullable"] and not getattr(bind_obj, key, None):
            if isinstance(prop_info["typename"], str):
                sub_obj = self._get_bindable_object(
                    model_factory(prop_info["typename"], self.version, self.appschema)
                )()
                setattr(bind_obj, key, [sub_obj] if prop_info["list"] else sub_obj)

        value = getattr(bind_obj, key, None)
        with (
            ui.list()
            .classes("w-full")
            .props("bordered separator dense")
            .mark("sub_form")
        ) as item_list:
            sub_model = None
            if isinstance(prop_info["typename"], str):
                sub_model = model_factory(
                    prop_info["typename"],
                    self.version,
                    self.appschema,
                )
                with ui.item().props("dense"):
                    LabelWithDescription(
                        sub_model.get_name(), sub_model.__doc__
                    ).classes("text-bold")
                ui.separator()
            if not preview:
                if prop_info["list"] or (prop_info["nullable"] and not value):
                    with ui.item().classes("justify-center"):
                        if isinstance(prop_info["typename"], str):
                            ui.button(
                                icon="add",
                                on_click=add_form,
                            ).bind_enabled_from(self, "editable").props("flat round")
                        else:
                            with (
                                ui.dropdown_button(
                                    text="Datentyp auswählen",
                                    icon="add",
                                    auto_close=True,
                                )
                                .bind_enabled_from(self, "editable")
                                .props("flat rounded no-caps")
                            ):
                                for type_name in prop_info["typename"]:
                                    ui.item(
                                        type_name,
                                        on_click=lambda x: add_form(
                                            next(x.sender.descendants()).text
                                        ),
                                    )
            if value is not None:
                if isinstance(value, list):
                    for index, item in enumerate(value):
                        await build_item(sub_model, item, index)
                else:
                    await build_item(sub_model, value)

    async def _build_model_form(
        self,
        model: type[BaseFeature],
        bind_obj: type,
        main_form: bool = False,
        preview: bool = False,
    ) -> None:
        """Construct a form for a given feature.

        This method is also called recursively for nested data types.

        Args:
            model (BaseFeature): The BaseFeature class.
            bind_obj (type): The bindable class to bind input to.
            main_form (bool, optional): Whether it's the initial/parent form or a sub-form. Defaults to False.
            preview (bool, optional): Whether the form should be rendered in preview mode. Defaults to False.
        """
        if main_form:
            with ui.row(align_items="center") as class_info:
                LabelWithDescription(model.get_name(), model.__doc__).classes("text-h6")
            if not preview:
                with class_info:
                    with ui.row(align_items="center"):
                        row_filter = (
                            ui.input(
                                "Attributnamen und -definition filtern",
                                on_change=lambda x: (
                                    radio_filter.set_value("all")
                                    if x.value
                                    else radio_filter.set_value(radio_filter.value)
                                ),
                            )
                            .classes("on-right")
                            .style("width: 380px;")
                            .props("clearable square filled dense bg-color=white")
                        )
                        self.row_filter = row_filter
                        with row_filter.add_slot("prepend"):
                            ui.icon("o_filter_alt")
                        radio_filter = (
                            ui.radio(
                                {
                                    "all": "Alle Attribute",
                                    "mandatory": "Pflichattribute",
                                    "set": "Belegte Attribute",
                                },
                                value="all",
                                on_change=self._filter_properties,
                            )
                            .props("dense inline dark color=white")
                            .classes("on-right")
                        )
                        self.radio_filter = radio_filter
            class_info.move(self.header if not preview else self.parent, 0)
        with ui.grid(columns="minmax(0, 1fr) minmax(0, 2fr)").classes("gap-5"):
            for key, field_info in model.model_fields.items():
                if key == model.get_geom_field():
                    continue
                prop_info = model.get_property_info(key)
                if preview and (
                    prop_info["stereotype"] == "Association"
                    or getattr(bind_obj, key, None) is None
                ):
                    continue
                default = None
                if isinstance(
                    field_default := getattr(field_info, "default", None),
                    (float, int, str),
                ):
                    default = field_default
                with ui.row().classes("w-full min-w-0 pr-8") as property_row:
                    if main_form and not preview:
                        property_row.bind_visibility_from(
                            row_filter,
                            "value",
                            backward=lambda filter_value,
                            key=key,
                            docs=field_info.description: (
                                re.search(filter_value, key + str(docs), re.I)
                                if filter_value
                                else True
                            ),
                        )
                        marker = ["property_row", key]
                        if prop_info["nullable"]:
                            marker.append("nullable")
                        property_row.mark(" ".join(marker))
                    LabelWithDescription(
                        key,
                        "Identifikator des Features."
                        if key == "id"
                        else field_info.description,
                    ).classes("text-bold")
                with (
                    ui.row()
                    .style("min-width: 250px;")
                    .classes("w-full min-w-0") as property_row
                ):
                    if main_form and not preview:
                        property_row.bind_visibility_from(
                            row_filter,
                            "value",
                            backward=lambda filter_value,
                            key=key,
                            docs=field_info.description: (
                                re.search(filter_value, key + str(docs), re.I)
                                if filter_value
                                else True
                            ),
                        )
                        marker = ["property_row", key]
                        if prop_info["nullable"]:
                            marker.append("nullable")
                        property_row.mark(" ".join(marker))
                    if key == "id":
                        input = (
                            ui.label()
                            .bind_text_from(
                                bind_obj,
                                key,
                            )
                            .classes("w-full")
                        )
                        continue
                    if key == "stylesheetId":
                        input = (
                            ui.label()
                            .bind_text_from(
                                bind_obj,
                                key,
                                backward=lambda x: (
                                    x.split("/")[-1]
                                    if x and "/" in x
                                    else "keine passende ID gefunden"
                                ),
                            )
                            .classes("w-full")
                        )
                        continue
                    match prop_info["stereotype"]:
                        case "BasicType":
                            match prop_info["typename"]:
                                case "CharacterString":
                                    if (
                                        key == "art"
                                        and "stylesheetId" in model.model_fields.keys()
                                    ):
                                        options = {}
                                        if existing := getattr(bind_obj, key, None):
                                            options.update(
                                                {
                                                    item: f"{item} (?)"
                                                    for item in existing
                                                }
                                            )
                                        input = (
                                            ui.select(
                                                options,
                                                label=prop_info["typename"],
                                                on_change=self._set_stylesheetId_and_schriftinhalt,
                                                multiple=prop_info["list"],
                                                clearable=prop_info["nullable"],
                                                validation=None
                                                if prop_info["nullable"]
                                                else {"Pflichtfeld": lambda x: x},
                                            )
                                            .bind_value(
                                                bind_obj,
                                                key,
                                                forward=lambda x: None if not x else x,
                                            )
                                            .bind_enabled_from(self, "editable")
                                            .props("stack-label square filled dense")
                                            .classes("w-full")
                                            .mark("art")
                                        )
                                        self.art_input = input
                                        self._get_art_options()
                                        input.validate()
                                    elif prop_info["list"]:
                                        ui.label("TODO")
                                    else:
                                        input = (
                                            ui.textarea(
                                                prop_info["typename"],
                                                validation=None
                                                if prop_info["nullable"]
                                                else {"Pflichtfeld": lambda x: x},
                                            )
                                            .bind_value(
                                                bind_obj,
                                                key,
                                                forward=lambda x: None if not x else x,
                                            )
                                            .props(
                                                "stack-label clearable autogrow square filled dense"
                                            )
                                            .classes("w-full q-pa-none q-gutter-none")
                                        )
                                        input.bind_enabled_from(
                                            self,
                                            "editable",
                                            backward=lambda v, input=input: input.props(
                                                **{
                                                    "add"
                                                    if not v
                                                    else "remove": "readonly"
                                                }
                                            ),
                                        )
                                        input.validate()
                                case "Integer":
                                    if prop_info["list"]:
                                        ui.label("TODO")
                                    else:
                                        input = (
                                            ui.number(
                                                prop_info["typename"],
                                                value=default,
                                                precision=0,
                                                format="%d",
                                                validation=None
                                                if prop_info["nullable"]
                                                else {
                                                    "Pflichtfeld": lambda x: (
                                                        x is not None
                                                    )
                                                },
                                            )
                                            .bind_value(
                                                bind_obj,
                                                key,
                                                forward=lambda x: (
                                                    int(x) if x is not None else None
                                                ),
                                                # backward=lambda x: x,
                                            )
                                            .props(
                                                "stack-label clearable square filled dense"
                                            )
                                            .classes("w-full q-pa-none q-gutter-none")
                                        )
                                        input.bind_enabled_from(
                                            self,
                                            "editable",
                                            backward=lambda v, input=input: input.props(
                                                **{
                                                    "add"
                                                    if not v
                                                    else "remove": "readonly"
                                                }
                                            ),
                                        )
                                        input.validate()
                                case "Decimal":
                                    if prop_info["list"]:
                                        ui.label("TODO")
                                    else:
                                        input = (
                                            ui.number(
                                                prop_info["typename"],
                                                # value=default,
                                                validation=None
                                                if prop_info["nullable"]
                                                else {
                                                    "Pflichtfeld": lambda x: (
                                                        x is not None
                                                    )
                                                },
                                            )
                                            .bind_value(
                                                bind_obj,
                                                key,
                                                # backward=lambda x: None if not x else x,
                                            )
                                            .props(
                                                "stack-label clearable square filled dense"
                                            )
                                            .classes("w-full q-pa-none q-gutter-none")
                                        )
                                        input.bind_enabled_from(
                                            self,
                                            "editable",
                                            backward=lambda v, input=input: input.props(
                                                **{
                                                    "add"
                                                    if not v
                                                    else "remove": "readonly"
                                                }
                                            ),
                                        )
                                        input.validate()
                                case "Boolean":
                                    ui.switch(value=default).props(
                                        "toggle-indeterminate"
                                    ).bind_value(bind_obj, key).bind_enabled_from(
                                        self, "editable"
                                    )
                                case "URI":
                                    if prop_info["list"]:
                                        ui.label("TODO")
                                    else:
                                        input = (
                                            DocumentSelect(
                                                bind_obj,
                                                key,
                                                label=prop_info["typename"],
                                                validation={
                                                    "Keine URI": self.validate_uri
                                                }
                                                if prop_info["nullable"]
                                                else {
                                                    "Pflichtfeld": lambda x: x,
                                                    "Keine URI": self.validate_uri,
                                                },
                                            )
                                            .classes("w-full q-pa-none q-gutter-none")
                                            .props("stack-label square filled dense")
                                        )
                                        input.bind_enabled_from(
                                            self,
                                            "editable",
                                            backward=lambda v, input=input: input.props(
                                                **{
                                                    "add"
                                                    if not v
                                                    else "remove": "readonly"
                                                }
                                            ),
                                        )
                                        input.validate()
                                case _:
                                    ui.label("BasicType")
                        case "Enumeration":
                            input = (
                                ui.select(
                                    {
                                        k: f"{v['name']} ({k})"
                                        for k, v in prop_info["enum_info"].items()
                                    },
                                    label=prop_info["typename"],
                                    multiple=prop_info["list"],
                                    clearable=prop_info["nullable"],
                                    validation=None
                                    if prop_info["nullable"]
                                    else {"Pflichtfeld": lambda x: x},
                                )
                                .classes("w-full q-pa-none q-gutter-none")
                                .props("stack-label square filled dense")
                                .bind_value(
                                    bind_obj,
                                    key,
                                    backward=lambda x: None if not x else x,
                                )
                            )
                            input.bind_enabled_from(
                                self,
                                "editable",
                                backward=lambda v, input=input: input.props(
                                    **{"add" if not v else "remove": "readonly"}
                                ),
                            )
                            if prop_info["list"]:
                                input.props("use-chips")
                            input.validate()
                        case "Codelist":
                            input = (
                                ui.select(
                                    await get_codelist_options(
                                        values=getattr(bind_obj, key, []),
                                        codelist=prop_info["typename"],
                                    ),
                                    label=prop_info["typename"],
                                    multiple=prop_info["list"],
                                    clearable=prop_info["nullable"],
                                    with_input=True,
                                    # new_value_mode="add-unique",
                                    # key_generator=lambda value,
                                    # codelist=prop_info[
                                    #     "typename"
                                    # ]: f"urn:{self.appschema}:codelist:{codelist}:{value}",
                                    validation={
                                        "Keine URI": lambda value: (
                                            self.validate_uri(value)
                                            if isinstance(value, str)
                                            else all(
                                                [
                                                    self.validate_uri(item)
                                                    for item in value
                                                ]
                                            )
                                            if value
                                            else True
                                        )
                                    }
                                    if prop_info["nullable"]
                                    else {
                                        "Pflichtfeld": lambda x: x,
                                        "Keine URI": lambda value: (
                                            self.validate_uri(value)
                                            if isinstance(value, str)
                                            else all(
                                                [
                                                    self.validate_uri(item)
                                                    for item in value
                                                ]
                                            )
                                            if value
                                            else True
                                        ),
                                    },
                                )
                                .bind_value(
                                    bind_obj,
                                    key,
                                    forward=lambda x: None if not x else x,
                                )
                                .classes("w-full q-pa-none q-gutter-none")
                                .props("stack-label square filled dense")
                            )
                            input.bind_enabled_from(
                                self,
                                "editable",
                                backward=lambda v, input=input: input.props(
                                    **{"add" if not v else "remove": "readonly"}
                                ),
                            )
                            if prop_info["list"]:
                                input.props("use-chips")
                            input.validate()
                            with input.add_slot("after"):
                                ui.button(
                                    icon="o_open_in_new",
                                    on_click=lambda input=input: ui.navigate.to(
                                        input.value
                                    ),
                                ).bind_enabled_from(
                                    input,
                                    "value",
                                    backward=lambda value: (
                                        "registry.gdi-de.org" in value
                                        if value
                                        else False
                                    ),
                                ).props("square unelevated")
                        case "Temporal":
                            match prop_info["typename"]:
                                case "Date":
                                    property_row.tooltip("Eingabe über Kalender")
                                    with (
                                        ui.input(
                                            label=prop_info["typename"],
                                            validation=None
                                            if prop_info["nullable"]
                                            else {
                                                "Pflichtfeld": lambda x: x,
                                            },
                                        )
                                        .bind_value_from(
                                            bind_obj,
                                            key,
                                            backward=lambda x: (
                                                ", ".join(x)
                                                if isinstance(x, list)
                                                else x
                                            ),
                                        )
                                        .classes("w-full q-pa-none q-gutter-none")
                                        .style("pointer-events: none;")
                                        .props(
                                            """
                                            stack-label
                                            square
                                            dense
                                            filled
                                            """
                                        ) as date_viewer
                                    ):
                                        with ui.dialog() as menu:
                                            date_picker = (
                                                ui.date()
                                                .bind_value(
                                                    bind_obj,
                                                    key,
                                                    forward=lambda x: (
                                                        sorted(x)
                                                        if isinstance(x, list)
                                                        else x or None
                                                    ),
                                                )
                                                .bind_enabled_from(self, "editable")
                                                .classes("w-96")
                                                .props(
                                                    f"""
                                                    title='{key}'
                                                    :locale='{
                                                        orjson.dumps(
                                                            {
                                                                "daysShort": [
                                                                    "So",
                                                                    "Mo",
                                                                    "Di",
                                                                    "Mi",
                                                                    "Do",
                                                                    "Fr",
                                                                    "Sa",
                                                                ],
                                                                "months": [
                                                                    "Januar",
                                                                    "Februar",
                                                                    "März",
                                                                    "April",
                                                                    "Mai",
                                                                    "Juni",
                                                                    "Juli",
                                                                    "August",
                                                                    "September",
                                                                    "Oktober",
                                                                    "November",
                                                                    "Dezember",
                                                                ],
                                                                "monthsShort": [
                                                                    "Jan",
                                                                    "Feb",
                                                                    "Mär",
                                                                    "Apr",
                                                                    "Mai",
                                                                    "Jun",
                                                                    "Jul",
                                                                    "Aug",
                                                                    "Sep",
                                                                    "Okt",
                                                                    "Nov",
                                                                    "Dez",
                                                                ],
                                                            }
                                                        ).decode()
                                                    }'
                                                    """
                                                )
                                            )
                                            if prop_info["list"]:
                                                date_picker.props(
                                                    "multiple subtitle='Liste von Datumsangaben'"
                                                )
                                            else:
                                                date_picker.props(
                                                    "subtitle='Datumsangabe'"
                                                )
                                                date_picker.on_value_change(menu.close)
                                        with date_viewer.add_slot("append"):
                                            ui.icon("edit_calendar").on(
                                                "click", menu.open
                                            ).classes("cursor-pointer").style(
                                                "pointer-events: auto;"
                                            )
                                        date_viewer.bind_enabled_from(
                                            self,
                                            "editable",
                                            backward=lambda v, input=date_viewer: (
                                                input.props(
                                                    **{
                                                        "add"
                                                        if not v
                                                        else "remove": "readonly"
                                                    }
                                                )
                                            ),
                                        )
                                        date_viewer.validate()
                                case "TM_Duration":
                                    if not prop_info["list"]:
                                        input = (
                                            ui.number(
                                                prop_info["typename"],
                                                value=default,
                                                suffix="Tage",
                                                step=1,
                                                precision=0,
                                                format="%d",
                                                validation=None
                                                if prop_info["nullable"]
                                                else {
                                                    "Pflichtfeld": lambda x: (
                                                        x is not None
                                                    )
                                                },
                                            )
                                            .bind_value(
                                                bind_obj,
                                                key,
                                                forward=lambda x: (
                                                    x * 86400 if x else None
                                                ),
                                                backward=lambda x: (
                                                    x / 86400 if x else None
                                                ),
                                            )
                                            .props(
                                                "stack-label clearable square filled dense"
                                            )
                                            .classes("w-full q-pa-none q-gutter-none")
                                        )
                                        input.bind_enabled_from(
                                            self,
                                            "editable",
                                            backward=lambda v, input=input: input.props(
                                                **{
                                                    "add"
                                                    if not v
                                                    else "remove": "readonly"
                                                }
                                            ),
                                        )
                                        input.validate()
                                    else:
                                        ui.label("TODO")
                                case _:
                                    ui.label("TODO Temporal")
                        case "DataType":
                            await self._build_sub_form(
                                key, bind_obj, prop_info, preview
                            )
                        case "Association":
                            if prop_info["list"]:
                                input = AssociationList(bind_obj, key)
                            else:
                                input = (
                                    ui.select(
                                        [getattr(bind_obj, key)],
                                        multiple=prop_info["list"],
                                        clearable=prop_info["nullable"],
                                        validation=None
                                        if prop_info["nullable"]
                                        else {"Pflichtfeld": lambda x: x},
                                    )
                                    .classes("w-full q-pa-none q-gutter-none")
                                    .props("square filled dense")
                                    .bind_value(
                                        bind_obj,
                                        key,
                                        backward=lambda x: None if not x else x,
                                    )
                                )
                                targets = await run.io_bound(
                                    self._get_association_targets,
                                    key,
                                    prop_info["typename"],
                                    [],
                                    app.storage.client.get("plan_id", None),
                                )
                                if not targets:
                                    input.set_label(
                                        "keine passenden Features vorhanden"
                                    )
                                else:
                                    input.set_options(targets)
                                if not input.value:  # only if no value was set
                                    if parent_id := app.storage.client.get(
                                        "parent_id", None
                                    ):
                                        input.set_value(parent_id)
                                    elif len((keys := list(targets.keys()))) == 1:
                                        input.set_value(keys[0])
                                input.bind_enabled_from(
                                    self,
                                    "editable",
                                    backward=lambda v, input=input: input.props(
                                        **{"add" if not v else "remove": "readonly"}
                                    ),
                                )
                                input.validate()
                                with input.add_slot("after"):
                                    ui.button(
                                        icon="o_open_in_new",
                                        on_click=lambda input=input: ui.navigate.to(
                                            f"/feature/{input.value}"
                                        ),
                                    ).bind_enabled_from(input, "value").props(
                                        "square unelevated"
                                    )
                                    ui.button(
                                        icon="o_preview",
                                        on_click=lambda input=input: (
                                            self._preview_association(input.value)
                                        ),
                                    ).bind_enabled_from(input, "value").props(
                                        "square unelevated"
                                    )
                        case "Measure":
                            uom = str(prop_info["uom"])
                            initial_value = isinstance(
                                getattr(bind_obj, key, None), dict
                            )
                            input = (
                                ui.number(
                                    prop_info["typename"],
                                    value=getattr(bind_obj, key)["value"]
                                    if initial_value
                                    else None,
                                    min=-360.0 if uom == "grad" else None,
                                    max=360.0 if uom == "grad" else None,
                                    validation=None
                                    if prop_info["nullable"]
                                    else {"Pflichtfeld": lambda x: x},
                                )
                                .classes("w-full q-pa-none q-gutter-none")
                                .props("stack-label square filled dense")
                            )
                            input.bind_enabled_from(
                                self,
                                "editable",
                                backward=lambda v, input=input: input.props(
                                    **{"add" if not v else "remove": "readonly"}
                                ),
                            )
                            with input.add_slot("after"):
                                uom_input = (
                                    ui.input(
                                        "uom",
                                        value=getattr(bind_obj, key)["uom"]
                                        if initial_value
                                        else uom,
                                    )
                                    .classes("w-24 q-pa-none q-gutter-none")
                                    .props("stack-label square filled dense")
                                ).tooltip(
                                    "Vorsicht beim Editieren - in den meisten Fällen ist die Maßeinheit vom Standard vorgegeben"
                                )
                                input.bind_value(
                                    bind_obj,
                                    key,
                                    forward=lambda x, uom_input=uom_input: (
                                        None
                                        if not x
                                        else {"value": x, "uom": uom_input.value}
                                    ),
                                    backward=lambda x: None if not x else x["value"],
                                )
                                uom_input.bind_enabled_from(
                                    self,
                                    "editable",
                                    backward=lambda v, input=uom_input: input.props(
                                        **{"add" if not v else "remove": "readonly"}
                                    ),
                                ).bind_value(
                                    bind_obj,
                                    key,
                                    forward=lambda x, input=input: (
                                        None
                                        if not input.value
                                        else {"value": input.value, "uom": x}
                                    ),
                                    backward=lambda x, uom=uom: (
                                        uom if not x else x["uom"]
                                    ),
                                )
                            input.validate()
                        case _:
                            ui.label("TODO")
        self.rendered = True

    async def render_form(
        self,
        feature_type: str,
        feature: dict,
        main_form: bool = True,
        preview: bool = False,
    ):
        """Render the form for a given feature.

        Args:
            feature_type (str): The type of the feature.
            feature (dict): The feature data as a dictionary.
            main_form (bool, optional): Whether this is the main form. Defaults to True.
            preview (bool, optional): Whether to render in preview mode. Defaults to False.
        """
        self.parent.clear()
        self.featuretype = feature_type
        self.model = model_factory(feature_type, self.version, self.appschema)
        if main_form:
            if (geom := feature.pop("geometry", None)) and (
                model_geom := self.model.get_geom_field()
            ):
                feature[model_geom] = geom
        self.feature = self._get_bindable_object(self.model)(**feature)
        await self._build_model_form(self.model, self.feature, main_form, preview)
        self._propagate_data()



"""Remove unique index from doc filename.

Allow files with the same name for different plans.

Revision ID: 5bc8757a75a8
Revises: 53a0d4789886
Create Date: 2026-03-02 19:59:00.908118

"""

from typing import Sequence, Union

from alembic import op
from xplan_tools.settings import get_settings

# revision identifiers, used by Alembic.
revision: str = "5bc8757a75a8"
down_revision: Union[str, Sequence[str], None] = "53a0d4789886"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

schema = get_settings().db_schema or "public"


def upgrade() -> None:
    """Upgrade schema."""
    op.drop_index(
        op.f("ix_docs_filename"), table_name="docs", schema=schema, if_exists=True
    )
    op.create_index(
        op.f("ix_docs_filename"),
        "docs",
        ["filename"],
        schema=schema,
    )


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_index(
        op.f("ix_docs_filename"), table_name="docs", schema=schema, if_exists=True
    )
    op.create_index(
        op.f("ix_docs_filename"), "docs", ["filename"], schema=schema, unique=True
    )



"""Add table to store binary data.

Revision ID: 53a0d4789886
Revises: 3715fd158e35
Create Date: 2026-02-23 09:52:42.731689
"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from xplan_tools.settings import get_settings

# revision identifiers, used by Alembic.
revision: str = "53a0d4789886"
down_revision: Union[str, Sequence[str], None] = "3715fd158e35"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

schema = get_settings().db_schema or "public"


def upgrade() -> None:
    """Upgrade schema."""
    op.create_table(
        "docs",
        sa.Column("pk", sa.BigInteger(), sa.Identity(always=True), nullable=False),
        sa.Column(
            "id",
            sa.Uuid(as_uuid=False),
            nullable=False,
        ),
        sa.Column("filename", sa.Text(), nullable=False),
        sa.Column("filetype", sa.Text(), nullable=False),
        sa.Column("filedata", postgresql.BYTEA(), nullable=False),
        sa.ForeignKeyConstraint(
            ["id"],
            [f"{schema}.coretable.id"],
            ondelete="CASCADE",
            initially="DEFERRED",
            deferrable=True,
        ),
        sa.PrimaryKeyConstraint("pk"),
        schema=schema,
    )
    op.create_index(
        op.f("ix_docs_filename"), "docs", ["filename"], schema=schema, unique=True
    )
    op.create_index(op.f("ix_docs_id"), "docs", ["id"], schema=schema)


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_index(op.f("ix_docs_id"), table_name="docs", schema=schema)
    op.drop_index(
        op.f("ix_docs_filename"), table_name="docs", schema=schema, if_exists=True
    )
    op.drop_table("docs", schema=schema)



import logging

from fastapi import Request
from nicegui import ui

from xmas_app.components.planmigration import PlanMigration
from xmas_app.components.plantree import PlanTree
from xmas_app.pages.base import BasePage
from xmas_app.util.webchannel import connect_signal, initialize_qwebchannel

logger = logging.getLogger("xmas_app")


class PlanToolsPage(BasePage):
    """Class to encapsulate page content and methods manipulate a plan.

    Displays plans from the database in a table view and allow to load/download/delete them.
    Allows initialization of the digitalization of new plans in QGIS.
    Provides an Uploader to import GML instance files of supported appschemas.
    """

    def __init__(
        self,
        request: Request,
        plan_id: str,
    ):
        """Initializes the PlanManagerPage.

        Content is rendered via the `render()` method.

        Args:
            request: a `fastapi.Request` for e.g. header extraction
            appschema: the appschema for new plans requested by the client
            version: the appschema version for new plans requested by the client
        """
        super().__init__(request)
        self.plan_id = plan_id

    async def render(self):
        """Renders the page content."""
        await ui.context.client.connected()

        if self.client_is_qgis:
            await initialize_qwebchannel()
            await connect_signal("featureAdded")

        with ui.tabs().classes("w-full").props("no-caps dense") as tabs:
            ui.tab("tree", "Featurebaum", "schema").classes("w-full")
            ui.tab("migrate", "Versionsmigration", "upgrade").classes("w-full")

        self.tab_panels = ui.tab_panels(tabs, value="tree").classes("w-full")
        with self.tab_panels:
            with ui.tab_panel("tree"):
                await PlanTree(self.plan_id).render()
            with ui.tab_panel("migrate"):
                await PlanMigration(self.plan_id).render()



import os
from functools import cached_property, lru_cache
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Literal, Optional, Self

from alembic import command
from alembic.config import Config
from pydantic import HttpUrl, model_validator
from pydantic_extra_types.semantic_version import SemanticVersion
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)
from xplan_tools.interface.db import AsyncDBRepository, DBRepository
from xplan_tools.model.base import Appschema

from xmas_app.models.mappings import AssociationTableMapping


class Settings(BaseSettings):
    debug: bool = False

    PGUSER: Optional[str] = None
    PGPASSWORD: Optional[str] = None
    PGHOST: Optional[str] = None
    PGPORT: Optional[str] = None
    PGDATABASE: Optional[str] = None
    PGSERVICE: Optional[str] = None

    appschema: str
    appschema_version: str
    appschema_cls: Appschema | None = None
    app_port: int
    db_type: str = "postgres"  # defaulting to postgres
    app_mode: Literal["dev", "prod"] = "prod"
    codelist_repo: HttpUrl = HttpUrl(
        "https://registry.gdi-de.org/codelist/de.xleitstelle.xplanung"
    )
    qgis_plugin_name: str = "XMAS-Plugin"
    qgis_plugin_min_version: SemanticVersion = SemanticVersion(major=0, minor=19)
    mappings_toml: str = str(Path(__file__).parent / "resources/mappings.toml")
    generate_schriftinhalt: bool = True
    static_file_dir: Path = Path(mkdtemp(prefix="xmas_"))

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8")

    @model_validator(mode="after")
    def _after_validator(self) -> Self:
        self.appschema_cls = Appschema.from_prefix(
            self.appschema, self.appschema_version
        )
        return self

    def model_post_init(self, context: Any) -> None:
        """Post-initialization steps."""
        os.environ.update(
            self.model_dump(
                include={
                    "PGUSER",
                    "PGHOST",
                    "PGPASSWORD",
                    "PGPORT",
                    "PGDATABASE",
                    "PGSERVICE",
                },
                exclude_none=True,
            )
        )
        self.static_file_dir.mkdir(exist_ok=True)

    @cached_property
    def repo(self):
        """Initialize and return repo."""
        return DBRepository(
            "postgresql://",
        )

    @cached_property
    def async_repo(self):
        """Initialize and return async repo."""
        return AsyncDBRepository(
            "postgresql://",
        )


@lru_cache
def get_settings() -> Settings:
    return Settings()


class Mappings(BaseSettings):
    association_table: AssociationTableMapping

    model_config = SettingsConfigDict(toml_file=get_settings().mappings_toml)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (TomlConfigSettingsSource(settings_cls),)


@lru_cache
def get_mappings() -> Mappings:
    return Mappings()


def initialize_db() -> None:
    """Ensure DB repository is initialized and runs app-specific alembic migrations."""
    repo = get_settings().repo
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", "xmas_app:models:migrations")
    with repo._engine.connect() as connection:
        alembic_cfg.attributes["connection"] = connection
        command.upgrade(alembic_cfg, "head")



import json
import logging
import re

from nicegui import run, ui
from nicegui.events import (
    ValueChangeEventArguments,
)
from xplan_tools.model import model_factory

from xmas_app.components.buttons import FeatureInteractionButton
from xmas_app.settings import get_settings
from xmas_app.util.db import (
    get_nodes,
)

logger = logging.getLogger("xmas_app")


class PlanTree(ui.column):
    def __init__(self, plan_id: str) -> None:
        super().__init__()
        self.plan_id = plan_id
        ui.on("featureAdded", self._update_nodes)

    async def render(self):
        self.action_dialog = ui.dialog().on("hide", lambda e: e.sender.clear())

        tree_filter = (
            ui.input("Baum filtern")
            .props("clearable square filled dense")
            .classes("w-full")
        )
        with tree_filter.add_slot("after"):
            ui.button(icon="refresh", on_click=self._update_nodes).tooltip(
                "Baum aktualisieren"
            )
        ui.separator()

        self.spinner_box = ui.row().classes("w-full justify-center items-center")
        with self.spinner_box:
            ui.spinner(size="xl")

        result = await run.io_bound(get_nodes, self.plan_id)
        if result:
            nodes, featuretype, id, appschema = result
        else:
            return
        self.nodes = nodes

        self.tree = ui.tree(
            self.nodes,
            on_select=self._show_menu,
            tick_strategy=None,
        ).props("accordion no-transition")

        tree_filter.bind_value_to(self.tree, "filter")

        if appschema == "xplan":
            self.tree.expand([id, f"{featuretype}.{id}.bereich"])

        self.spinner_box.clear()

    async def _update_nodes(self):
        result = await run.io_bound(get_nodes, self.plan_id)
        if result:
            nodes, *_ = result
            self.nodes[:] = nodes
            self.tree.update()

    async def _show_menu(self, e: ValueChangeEventArguments):
        if e.value:
            path = e.value.split(".")
            if len(path) == 5:
                _, origin, _, target_featuretype, target = path
                feature = get_settings().repo.get(target)
                with self.action_dialog, ui.card():
                    ui.label(f"{target_featuretype} {target[:8]}").classes("text-bold")
                    FeatureInteractionButton(
                        "highlight_feature",
                        target,
                        geom=feature.get_geom_wkt() is not None,
                    )
                    FeatureInteractionButton("select_feature", target)
                    FeatureInteractionButton("show_attribute_form", target)
                    if not re.match("^.._Bereich$", target_featuretype):
                        ui.button(
                            "Feature löschen",
                            on_click=lambda _, target=target: self._on_delete(target),
                            icon="delete",
                        )
            elif len(path) == 3:
                origin_featuretype, origin_id, rel = path
                with self.action_dialog, ui.card(align_items="stretch"):
                    ui.label(origin_featuretype + " " + origin_id[:8]).classes(
                        "text-bold"
                    )
                    if rel == "bereich":
                        ui.label(f"Keine Aktionen für Referenz {repr(rel)} verfügbar")
                    else:
                        ui.label(f"Neues Feature für Referenz {repr(rel)} erstellen")
                        feature = get_settings().repo.get(origin_id)
                        prop_info = feature.get_property_info(rel)
                        typename = prop_info["typename"]
                        targets = typename if isinstance(typename, list) else [typename]
                        target_select = ui.select(
                            targets,
                            label="Featuretype auswählen",
                            value=None,
                            with_input=True,
                        )
                        if len(targets) == 1:
                            target_select.props("readonly")
                        with target_select.add_slot("append"):
                            ui.badge(str(len(targets))).props("floating")
                        with target_select.add_slot("after"):
                            dropdown = ui.dropdown_button(
                                icon="play_circle",
                                split=True,
                            )
                            target_select.on_value_change(
                                lambda e,
                                origin_featuretype=origin_featuretype,
                                origin_id=origin_id,
                                rel=rel: (
                                    self._prepare_add_feature(
                                        dropdown,
                                        origin_featuretype,
                                        origin_id,
                                        rel,
                                        e.value,
                                    )
                                )
                            )
                            if len(targets) == 1:
                                target_select.set_value(targets[0])
                            else:
                                dropdown.props("disable")

            else:
                return
            self.action_dialog.open()

    async def _prepare_add_feature(
        self,
        dropdown: ui.dropdown_button,
        origin_featuretype: str,
        origin_id: str,
        rel: str,
        target_featuretype: str,
    ):
        async def trigger_add_feature(geometry_type):
            rel_inv = origin_feature.get_property_info(rel)["assoc_info"]["reverse"]
            data = {
                "origin_featuretype": origin_featuretype,
                "origin_id": origin_id,
                "rel": rel,
                "rel_list": origin_feature.get_property_info(rel)["list"],
                "rel_inv": rel_inv,
                "rel_inv_list": target_model.get_property_info(rel_inv)["list"]
                if rel_inv
                else False,
                "target_featuretype": target_featuretype,
                "target_geometrytype": geometry_type,
            }
            logger.info("Sending add_feature to QWebChannel handler")
            payload_json = json.dumps(data)
            await ui.run_javascript(
                f"""
                    (window.ensureXmasHandler ? window.ensureXmasHandler() : Promise.reject('QWebChannel unavailable'))
                        .then(handler => handler.add_feature({payload_json}))
                        .catch(err => console.error('add_feature failed', err));
                """
            )

        GEOMETRY_MAP = {
            "Polygon": {"label": "Fläche", "icon": "o_space_dashboard"},
            "Line": {"label": "Linie", "icon": "show_chart"},
            "Point": {"label": "Punkt", "icon": "o_location_on"},
            "Null": {"label": "Keine Geometrie", "icon": "o_text_snippet"},
        }

        logger.debug("add_feature called")
        dropdown.clear()
        dropdown.props(remove="disable")
        origin_feature = get_settings().repo.get(origin_id)
        target_model = model_factory(
            target_featuretype,
            origin_feature.get_version(),
            origin_feature.get_appschema(),
        )
        geometry_types = []
        if geom_types := target_model.get_geom_types():
            for geom_type in geom_types:
                type_name = geom_type.get_name().replace("Multi", "")
                if type_name not in geometry_types:
                    geometry_types.append(type_name)
        else:
            geometry_types.append("Null")
        if len(geometry_types) == 1:
            dropdown.on_click(lambda: trigger_add_feature(geometry_types[0]))
            dropdown.props("disable-dropdown")
        else:
            with dropdown:
                dropdown.on_click(dropdown.open)
                dropdown.props(remove="disable-dropdown")
                for geometry_type in geometry_types:
                    with ui.item(
                        on_click=lambda geometry_type=geometry_type: (
                            trigger_add_feature(geometry_type)
                        ),
                    ):
                        with ui.item_section().props("avatar"):
                            ui.icon(GEOMETRY_MAP[geometry_type]["icon"])
                        with ui.item_section():
                            ui.item_label(GEOMETRY_MAP[geometry_type]["label"])

    async def _on_delete(self, target_id: str):
        with ui.dialog() as delete_dialog, ui.card(align_items="stretch"):
            ui.label("Soll das Objekt wirklich gelöscht werden?")
            ui.button(
                "Löschen", color="red", on_click=lambda: delete_dialog.submit(True)
            )
            ui.button("Abbrechen", on_click=lambda: delete_dialog.submit(False))

        result = await delete_dialog
        delete_dialog.delete()

        if result:
            try:
                get_settings().repo.delete(target_id)
                ui.notify(
                    f"Feature mit ID '{target_id}' wurde gelöscht", type="positive"
                )
                await self._update_nodes()
            except Exception as e:
                logger.exception("error deleting feature %r: %s", target_id, e)
                ui.notify(
                    f"Fehler beim Löschen vom Feature mit ID '{target_id}'",
                    type="negative",
                )



from uuid import UUID

from sqlalchemy import BigInteger, ForeignKey, Identity, Text
from sqlalchemy.dialects.postgresql import BYTEA
from sqlalchemy.orm import Mapped, mapped_column
from xplan_tools.model.orm import Base, Feature


class Document(Base):
    __tablename__ = "docs"
    __table_args__ = {"schema": Feature.__table_args__.get("schema")}
    pk: Mapped[int] = mapped_column(
        BigInteger,
        Identity(always=True),
        primary_key=True,
    )
    id: Mapped[UUID] = mapped_column(
        ForeignKey(
            Feature.id, ondelete="CASCADE", deferrable=True, initially="DEFERRED"
        ),
        index=True,
    )
    filename: Mapped[str] = mapped_column(Text, index=True)
    filetype: Mapped[str] = mapped_column(Text)
    filedata: Mapped[bytes] = mapped_column(BYTEA)



import logging
from typing import AsyncIterator
from zoneinfo import ZoneInfo

from sqlalchemy import Text, cast, delete, func, insert, literal, outerjoin, select
from sqlalchemy.dialects.postgresql import JSON as PG_JSON
from sqlalchemy.orm import defer, selectinload
from xplan_tools.model import model_factory
from xplan_tools.model.base import BaseCollection, BaseFeature
from xplan_tools.model.orm import Feature, Refs

from xmas_app.models.orm import Document
from xmas_app.settings import get_mappings, get_settings

LOCAL_TZ = ZoneInfo("Europe/Berlin")

logger = logging.getLogger("xmas_app")


async def get_db_plans(chunk_size: int = 500) -> AsyncIterator[list[dict]]:
    """Get plan data from the database.

    Query the database for plan objects, yielded in chunks.

    Args:
        chunk_size: Number of plans per chunk.

    Yields:
        list[dict]: Chunk of dicts containing plan data.
    """
    try:
        async with get_settings().async_repo.AsyncSession() as session:
            join_stmt = outerjoin(Feature, Document, Feature.id == Document.id)

            # Correlated subquery: aggregate bereiche as JSON array
            bereich = Feature.__table__.alias("bereich")
            bereiche_subq = (
                select(
                    func.coalesce(
                        func.json_agg(
                            func.json_build_object(
                                "id",
                                cast(Refs.related_id, Text),
                                "nummer",
                                bereich.c.properties["nummer"].as_string(),
                                "geometry",
                                bereich.c.geometry.isnot(None),
                            )
                        ),
                        cast(literal("[]"), PG_JSON),
                    )
                )
                .select_from(
                    outerjoin(Refs.__table__, bereich, Refs.related_id == bereich.c.id)
                )
                .where(Refs.base_id == Feature.id, Refs.rel == "bereich")
                .correlate(Feature)
                .scalar_subquery()
                .label("bereiche")
            )

            stmt = (
                select(
                    Feature,
                    func.count(Document.id).label("document_count"),
                    func.array_agg(Document.filename)
                    .filter(Document.filetype.startswith("image"))
                    .label("document_filenames"),
                    bereiche_subq,
                )
                .select_from(join_stmt)
                .where(Feature.featuretype.regexp_match("^.*Plan$"))
                .group_by(Feature.pk)
                .execution_options(yield_per=chunk_size)
            )
            result = await session.stream(stmt)
            chunk = []
            async for feature, doc_count, images, bereiche_data in result:
                plan_name = feature.properties["name"]
                updated_local = feature.updated_at.astimezone(LOCAL_TZ)
                created_local = feature.created_at.astimezone(LOCAL_TZ)

                chunk.append(
                    {
                        "id": str(feature.id),
                        "name": plan_name,
                        "featuretype": feature.featuretype,
                        "appschema": f"{feature.appschema.upper()} {feature.version}",
                        "updated": f"{updated_local.strftime('%d.%m.%Y %H:%M')}",
                        "created": f"{created_local.strftime('%d.%m.%Y %H:%M')}",
                        "bereiche": bereiche_data
                        if isinstance(bereiche_data, list)
                        else [],
                        "docs": doc_count,
                        "images": images or [],
                    }
                )
                if len(chunk) >= chunk_size:
                    yield chunk
                    chunk = []
            if chunk:
                yield chunk

    except Exception as e:
        logger.exception("Failed to get plan data: %r", e)
        raise


def get_feature_number(plan_id: str) -> int:
    """Return the number of features of a plan."""
    with (repo := get_settings().repo).Session() as session:
        feature_number = session.scalar(
            func.count(repo.get_feature_graph(plan_id, depth=2).c.id)
        )
    return feature_number


def get_nodes(id: str) -> tuple[list[dict], str, str, str] | None:
    """Build tree nodes for ui.tree element."""

    def build_node(feature: Feature, path: str) -> dict:
        features_set.add(feature.id)
        node = {
            "id": f"{path}.{feature.featuretype}.{feature.id}",
            "label": f"{feature.featuretype} {str(feature.id)[:8]}",
            "icon": get_icon(feature),
        }
        node["children"] = build_nodes(feature)
        return node

    def build_nodes(feature: Feature) -> list[dict]:
        model = model_factory(feature.featuretype, feature.version, feature.appschema)
        node_dict = {}
        for assoc in model.get_associations():
            prop_info = model.get_property_info(assoc)
            if prop_info["assoc_info"]["source_or_target"] == (
                "target"
                if (feature.featuretype == "FP_Plan" and assoc == "bereich")
                or (feature.featuretype == "FP_Bereich" and assoc == "gehoertZuPlan")
                else "source"
            ):
                continue
            node = {
                "id": f"{feature.featuretype}.{feature.id}.{assoc}",
                "label": assoc,
                "selectable": True,
                "icon": "link",
            }
            node_dict[assoc] = node
        for ref in feature.refs:
            path = ".".join(
                [
                    feature.featuretype,
                    str(feature.id),
                    ref.rel,
                ]
            )
            node = build_node(ref.feature_inv, path)
            node_dict[ref.rel].setdefault("children", []).append(node)

        nodes = []
        for node in node_dict.values():
            if children := node.get("children"):
                node["label"] += f" [{len(children)}]"
            nodes.append(node)
        return nodes

    def get_icon(feature: Feature):
        match feature.geometry_type:
            case "polygon":
                icon = "o_space_dashboard"
            case "line":
                icon = "show_chart"
            case "point":
                icon = "o_location_on"
            case _:
                icon = "o_text_snippet"
        return icon

    with get_settings().repo.Session() as session:
        stmt = (
            select(Feature)
            .options(
                selectinload(Feature.refs)
                .selectinload(Refs.feature_inv)
                .selectinload(Feature.refs)
                .selectinload(Refs.feature_inv)
                .selectinload(Feature.refs)
            )
            .where(Feature.id == id)
        )

        feature = session.execute(stmt).scalar_one_or_none()
        # feature = session.get(Feature, id)
        if not feature or "Plan" not in feature.featuretype:
            return

        features_set = {feature.id}

        tree_nodes = [
            {
                "id": feature.id,
                "label": feature.properties["name"],
                "selectable": False,
                "children": build_nodes(feature),
            }
        ]

        return (
            tree_nodes,
            feature.featuretype,
            feature.id,
            feature.appschema,
        )


def _feature_to_table_row(feature: Feature) -> dict:
    extra_property = getattr(
        get_mappings().association_table.extra_properties, feature.appschema
    ).get(feature.featuretype)
    extra_property_label = feature.properties.get(extra_property, "N/A")
    return {
        "id": str(feature.id),
        "featuretype": feature.featuretype,
        "geometry_type": feature.geometry_type,
        "updated": f"{feature.updated_at.strftime('%d.%m.%Y')} {feature.updated_at.strftime('%H:%M')}",
        "extra_property": f"{extra_property}={extra_property_label}"
        if extra_property
        else "N/A",
    }


def _rel_inv_is_list(feature: Feature, rel_inv: str | None) -> bool:
    # if there is no inverse relation, return True
    if not rel_inv:
        return True
    model = model_factory(feature.featuretype, feature.version, feature.appschema)
    return model.get_property_info(rel_inv)["list"]


def get_ref_objects(refs: list[str]) -> list[dict]:
    """Returns a list of data objects for existing feature references to use in table rows."""
    with get_settings().repo.Session() as session:
        stmt = select(Feature).where(Feature.id.in_(refs))
        features = session.execute(stmt).scalars().all()

    return [_feature_to_table_row(feature) for feature in features]


def get_ref_candidates(
    refs: list[str], plan_id: str, model: BaseFeature, rel: str
) -> list[dict]:
    """Returns a list of data objects for potential feature references to use in table rows."""
    prop_info = model.get_property_info(rel)
    typename = prop_info["typename"]
    featuretypes = typename if isinstance(typename, list) else [typename]
    rel_inv = model.get_property_info(rel)["assoc_info"]["reverse"]
    repo = get_settings().repo

    with repo.Session() as session:
        candidates = session.scalars(
            repo.get_feature_graph(plan_id).where(
                Feature.featuretype.in_(featuretypes) & Feature.id.not_in(refs)
            )
        ).all()
        return [
            _feature_to_table_row(feature)
            for feature in candidates
            if _rel_inv_is_list(feature, rel_inv)
        ]


def get_plan_documents(
    plan_id: str, details: bool = False, with_data: bool = False
) -> list[str | dict]:
    """Retrieves binary documents stored in the db for a given plan."""
    repo = get_settings().repo
    with repo.Session() as session:
        stmt = (
            select(Document)
            .where(Document.id == plan_id)
            .options(defer(Document.filedata))
        )
        docs = []
        for doc in session.scalars(stmt).all():
            if not details:
                docs.append(f"file:///{doc.filename}")
            else:
                doc_data = {
                    "pk": doc.pk,
                    "filename": doc.filename,
                    "filetype": doc.filetype,
                }
                if with_data:
                    doc_data["filedata"] = doc.filedata
                docs.append(doc_data)
        return docs


def get_doc_by_name(plan_id: str, filename: str) -> Document:
    """Retrieve a document by filename (case-insensitive)."""
    repo = get_settings().repo
    with repo.Session() as session:
        return session.scalar(
            select(Document)
            .where(Document.id == plan_id)
            .where(func.lower(Document.filename) == func.lower(filename))
        )


def save_documents(data: list[dict]) -> None:
    """Retrieves binary documents stored in the db for a given plan."""
    repo = get_settings().repo
    with repo.Session() as session:
        session.execute(insert(Document), data)
        session.commit()


def save_plan_and_documents(plan: BaseCollection, docs: list[dict]) -> None:
    """Save a plan collection with its documents in one transaction."""
    repo = get_settings().repo
    with repo.Session() as session:
        feature_list, refs_list, refs_inv_list = repo._serialize_bulk_features(plan)
        session.execute(insert(Feature), feature_list)
        refs_list.extend(
            [ref_inv for ref_inv in refs_inv_list if ref_inv not in refs_list]
        )
        if refs_list:
            session.execute(
                insert(Refs),
                refs_list,
            )
        if docs:
            session.execute(insert(Document), docs)
        session.commit()


def delete_documents(pks: list[int]) -> None:
    """Deletes documents from the db."""
    repo = get_settings().repo
    with repo.Session() as session:
        session.execute(delete(Document).where(Document.pk.in_(pks)))
        session.commit()



import io
import logging
import zipfile

from fastapi import Request
from nicegui import app, run, ui
from nicegui.events import (
    ClickEventArguments,
    GenericEventArguments,
    MultiUploadEventArguments,
)
from starlette.formparsers import MultiPartParser
from xplan_tools.interface.gml import GMLRepository
from xplan_tools.model.base import Appschema, BaseCollection

from xmas_app.components.documents import DocumentEditor
from xmas_app.pages.base import BasePage
from xmas_app.settings import get_settings
from xmas_app.util import get_model_for_select
from xmas_app.util.db import (
    get_db_plans,
    get_feature_number,
    get_plan_documents,
    save_plan_and_documents,
)

logger = logging.getLogger("xmas_app")

MultiPartParser.spool_max_size = 1024 * 1024 * 50


class PlanManagerPage(BasePage):
    """Class to encapsulate page content and methods to manage plans.

    Displays plans from the database in a table view and allow to load/download/delete them.
    Allows initialization of the digitalization of new plans in QGIS.
    Provides an Uploader to import GML instance files of supported appschemas.
    """

    def __init__(
        self,
        request: Request,
        appschema: str,
        version: str,
    ):
        """Initializes the PlanManagerPage.

        Content is rendered via the `render()` method.

        Args:
            request: a `fastapi.Request` for e.g. header extraction
            appschema: the appschema for new plans requested by the client
            version: the appschema version for new plans requested by the client
        """
        super().__init__(request)
        ui.on("planCreated", self.update_plans)
        self.appschema = appschema
        self.version = version

    async def render(self):
        """Renders the page content."""
        self.root = ui.grid(rows="auto auto auto").classes("w-full")
        with self.root:
            self.plans_dialog = ui.dialog().props("full-height")
            with ui.column(align_items="start"):
                with ui.row().classes("h-96 w-full"):
                    self.plans_table = self._create_table()
            with ui.column(align_items="start").classes("p-4 pt-2 border border-solid"):
                ui.label("Neuen Plan erstellen").classes("text-bold")
                appschema_cls = Appschema.from_prefix(self.appschema, self.version)
                self.new_plan_select = (
                    ui.select(
                        {},
                        label=f"{appschema_cls.full_name.split()[0]} {appschema_cls.version} Planart auswählen",
                    )
                    .bind_enabled_from(self, "client_is_qgis")
                    .classes("w-full")
                )
                with self.new_plan_select.add_slot("after"):
                    ui.button(
                        icon="play_circle",
                        on_click=self._on_new_plan,
                    ).bind_enabled_from(self.new_plan_select, "value").tooltip(
                        "Planerstellung starten"
                    )
            with ui.column(align_items="start").classes("relative"):
                uploader = (
                    ui.upload(
                        multiple=True,
                        max_total_size=50000000,
                        label="GML-Datei importieren",
                        on_rejected=lambda: ui.notify(
                            "Falscher Dateityp", type="negative"
                        ),
                    )
                    .props(
                        "accept='text/xml,application/gml+xml,.xml,.gml' flat square bordered"
                    )
                    .classes("w-full")
                )
                new_ids = (
                    ui.checkbox("IDs generieren")
                    .tooltip("beim Import neue Feature-IDs erzeugen")
                    .props("dark")
                    .classes("text-white text-bold absolute top-0 right-0 p-2 mr-32")
                )
                uploader.on_multi_upload(
                    lambda e: self._on_import_plans(e, new_ids.value)
                )
        try:
            await ui.context.client.connected(timeout=10)
        except TimeoutError:
            return
        await self.update_plans()
        if self.client_is_qgis:
            await get_model_for_select(
                self.new_plan_select,
                feature_regex="^.*Plan$",
                appschema=self.appschema,
                appschema_version=self.version,
            )
            self.new_plan_select.props(remove="loading")
            self.new_plan_select.set_value(self.new_plan_select.options[0])

    async def _on_import_plans(self, event: MultiUploadEventArguments, new_ids: bool):
        """Import uploaded plans to the database."""
        self.plans_table.props("loading")
        imported = []
        failed = []
        try:
            for index, content in enumerate(event.contents):
                data = io.BytesIO(content.read())
                name = event.names[index]
                try:
                    repo = GMLRepository(data)
                except Exception as e:
                    logger.exception("Uploaded file could not be read: %r", e)
                    failed.append(name)
                    continue
                try:
                    collection = await run.io_bound(repo.get_all)
                except Exception as e:
                    logger.exception("Failed to read plan from file: %r", e)
                    failed.append(name)
                try:
                    if new_ids:
                        collection = collection.make_copy()
                    await run.io_bound(get_settings().repo.save_all, collection)
                    imported.append(name)
                except Exception as e:
                    logger.exception("Failed to save plan to database: %r", e)
                    failed.append(name)
            if failed:
                ui.notify(f"Import gescheitert: {', '.join(failed)}", type="negative")
            if imported:
                ui.notify(f"Import erfolgreich: {', '.join(imported)}", type="positive")
                for client in app.clients("/plans"):
                    client.run_javascript("emitEvent('planCreated')")
        finally:
            self.plans_table.props(remove="loading")

    async def _on_load_plan(self, e: ClickEventArguments):
        """Load a plan in QGIS.

        Handles loading a selected plan and sends the plan data to the QWebChannel handler.
        """
        logger.debug("retrieving plan data to send to QGIS")
        e.sender.props(add="loading")
        try:
            plans_data = []
            for selected_plan in self.plans_table.selected:
                appschema, version = selected_plan["appschema"].split()
                plans_data.append(
                    {
                        "plan_id": selected_plan["id"],
                        "plan_name": selected_plan["name"],
                        "appschema": appschema.lower(),
                        "version": version,
                        "plan_type": selected_plan["featuretype"],
                        "bereiche": selected_plan["bereiche"],
                        "images": selected_plan["images"],
                    }
                )
            logger.debug(f"Sending plan data to QWebChannel handler: {plans_data}")
            ui.run_javascript(f"""new QWebChannel(qt.webChannelTransport, function (channel) {{
                                        channel.objects.handler.load_plan({plans_data})
                                }});""")
        finally:
            e.sender.props(remove="loading")

    async def _on_delete_plan(self, e: ClickEventArguments):
        """Delete plan using repository method from XPlan-Tools."""
        e.sender.props(add="loading")
        plans_to_delete = self.plans_table.selected
        with ui.dialog() as dialog, ui.card():
            with ui.list():
                ui.item_label("Löschen für folgende Pläne bestätigen:").props(
                    "header"
                ).classes("text-bold text-black")
                for plan in plans_to_delete:
                    with ui.item():
                        with ui.item_section().props("avatar"):
                            ui.icon("delete_forever", color="primary")
                        with ui.item_section():
                            ui.item_label(plan["name"])
                            ui.item_label(plan["id"]).props("caption")
            with ui.row().classes("w-full justify-around"):
                ui.button(icon="check", on_click=lambda: dialog.submit(True))
                ui.button(icon="close", on_click=lambda: dialog.submit(False))
        try:
            result = await dialog
            dialog.delete()
            if not result:
                return

            plan_data = []
            for plan in plans_to_delete:
                try:
                    await run.io_bound(
                        get_settings().repo.delete_plan_by_id, plan["id"]
                    )
                    logger.info("Plan %r deleted successfully", plan["id"])
                    logger.info(
                        "Sending plan delete information to QWebChannel handler"
                    )
                    plan_data.append({"id": plan["id"], "name": plan["name"]})
                except Exception:
                    logger.exception("Failed to delete plan %r", plan["id"])
                    ui.notify(
                        f"Fehler beim Löschen des Plans {plan['name']!r}.",
                        type="negative",
                    )
            ui.run_javascript(f"""new QWebChannel(qt.webChannelTransport, function (channel) {{
                                        channel.objects.handler.deleted_plans({plan_data})
                                }});""")

            for client in app.clients("/plans"):
                client.run_javascript("emitEvent('planCreated')")
        finally:
            e.sender.props(remove="loading")

    def _on_new_plan(self):
        plan_data = {
            "featuretype": self.new_plan_select.value,
        }
        ui.run_javascript(f"""new QWebChannel(qt.webChannelTransport, function (channel) {{
                                    channel.objects.handler.create_plan({plan_data})
                            }});""")

    async def _on_copy_plan(self, e: ClickEventArguments):
        """Delete plan using repository method from XPlan-Tools."""
        e.sender.props(add="loading")
        try:
            for plan in self.plans_table.selected:
                try:
                    repo = get_settings().repo
                    collection = await run.io_bound(repo.get_plan_by_id, plan["id"])
                    name = collection.features[plan["id"]].name
                    collection.features[plan["id"]].name = f"{name} Kopie"
                    collection_copy, id_map = collection.make_copy(with_id_map=True)
                    docs = get_plan_documents(plan["id"], details=True, with_data=True)
                    for doc in docs:
                        doc["id"] = id_map[plan["id"]]
                        doc.pop("pk")
                    await run.io_bound(save_plan_and_documents, collection_copy, docs)
                    logger.info("Copy of Plan %r created successfully", plan["id"])
                except Exception:
                    logger.exception("Failed to copy plan %r", plan["id"])
                    ui.notify(
                        f"Fehler beim Kopieren des Plans {plan['name']!r}.",
                        type="negative",
                    )
            await self.update_plans()
        finally:
            e.sender.props(remove="loading")

    async def _on_download_plan(self, button: ui.dropdown_button, archive=False):
        """Downloads plan as GML file or ZIP archive including plan documents."""
        button.close()
        button.props(add="loading")
        try:
            features = {}
            overall_appschema = None
            overall_version = None
            for selected_plan in self.plans_table.selected:
                try:
                    logger.debug(
                        "retreiving plan with id %r from db", selected_plan["id"]
                    )
                    appschema, version = selected_plan["appschema"].split()
                    if overall_appschema and overall_version:
                        if overall_appschema != appschema:
                            return ui.notify(
                                f"Heterogene Anwendungsschemas ({overall_appschema}, {appschema}) werden nicht unterstützt",
                                type="negative",
                            )
                        if overall_version != version:
                            return ui.notify(
                                f"Heterogene Versionen ({overall_version}, {version}) werden nicht unterstützt",
                                type="negative",
                            )
                    overall_appschema = appschema
                    overall_version = version
                    collection = await run.io_bound(
                        get_settings().repo.get_plan_by_id, selected_plan["id"]
                    )
                    srid = collection.srid
                    features.update(collection.features)
                except Exception as e:
                    logger.exception(
                        "Failed to retrieve plan %r: %r", selected_plan["id"], e
                    )
                    ui.notify(
                        f"Fehler beim Abruf von Plan {selected_plan['name']}, überspringe",
                        type="warning",
                    )
                    continue
            overall_collection = BaseCollection(
                features=features,
                srid=srid,
                version=overall_version,
                appschema=overall_appschema.lower(),
            )
            buffer = io.BytesIO()
            logger.debug("Serializing to GML via GMLRepository")
            GMLRepository(buffer).save_all(overall_collection)
            data = buffer.getvalue()
            logger.debug("Prepared download payload (%d bytes)", len(data))
            if not archive:
                return ui.download.content(
                    data,
                    filename="xplan.gml",
                    media_type="application/gml+xml",
                )
            zip_buffer = io.BytesIO()
            with zipfile.ZipFile(
                zip_buffer, mode="w", compression=zipfile.ZIP_DEFLATED
            ) as zf:
                zf.writestr(f"{overall_appschema.lower()}.gml", data)
                for selected_plan in self.plans_table.selected:
                    for doc in get_plan_documents(
                        selected_plan["id"], details=True, with_data=True
                    ):
                        zf.writestr(doc["filename"], doc["filedata"])
            return ui.download.content(
                zip_buffer.getvalue(),
                filename=f"{overall_appschema.lower()}.zip",
                media_type="application/zip",
            )

        except Exception as e:
            logger.exception("Failed to download plans: %r", e)
            ui.notify("Fehler beim Plandownload", type="negative")
        finally:
            button.props(remove="loading")

    async def update_plans(self):
        """Update plan data from database."""
        self.plans_table.props("loading")
        try:
            self.plans_table.rows.clear()
            self.plans_table.selected.clear()
            async for chunk in get_db_plans(500):
                self.plans_table.add_rows(chunk)
            self._update_row_select_label()
        except Exception as e:
            logger.exception("Fehler beim Laden der Pläne: %r", e)
            if (
                "could not connect" in str(e).lower()
                or "connection refused" in str(e).lower()
            ):
                ui.notify("Keine Verbindung zur Datenbank!", type="negative")
            else:
                ui.notify(f"Fehler beim Laden der Pläne: {e}", type="negative")
        finally:
            self.plans_table.props(remove="loading")

    def _update_row_select_label(self):
        plan_number = len(self.plans_table.rows)
        self.plans_table.props(
            f":selected-rows-label='(numberOfRows) => `${{ numberOfRows }} von {plan_number} {'Plan' if plan_number == 1 else 'Plänen'} selektiert`'"
        )

    def _create_table(self) -> ui.table:
        """Creates a table with adequate columns, props and event handlers."""
        columns = [
            {
                "name": "name",
                "label": "Name",
                "field": "name",
                "align": "left",
                "sortable": True,
                "style": "width:50%;max-width:0;overflow:hidden;white-space:nowrap;text-overflow:ellipsis;",
            },
            {
                "name": "featuretype",
                "label": "Objektart",
                "field": "featuretype",
                "align": "left",
                "sortable": True,
            },
            {
                "name": "appschema",
                "label": "Appschema",
                "field": "appschema",
                "align": "left",
                "sortable": True,
            },
            {
                "name": "id",
                "label": "ID",
                "field": "id",
                "align": "left",
                "sortable": True,
                "classes": "hidden",
                "headerClasses": "hidden",
            },
            {
                "name": "updated",
                "label": "Aktualisierung",
                "field": "updated",
                "align": "left",
                "sortable": True,
                "classes": "hidden",
                "headerClasses": "hidden",
            },
            {
                "name": "created",
                "label": "Erstellung",
                "field": "created",
                "align": "left",
                "sortable": True,
                "classes": "hidden",
                "headerClasses": "hidden",
            },
            {
                "name": "docs",
                "label": "Dokumente",
                "field": "docs",
                "align": "left",
                "sortable": True,
                "classes": "hidden",
                "headerClasses": "hidden",
            },
        ]
        table = (
            ui.table(
                rows=[],
                columns=columns,
                selection="multiple",
            )
            .props(
                """
                no-data-label='keine passenden Features vorhanden'
                no-results-label='keine passenden Ergebnisse gefunden'
                flat
                square
                dense
                virtual-scroll
                bordered
            """
            )
            .classes("w-full h-full")
        )
        table.on("row-click", self._on_row_click)
        with table.add_slot("top"):
            with ui.column().classes("w-full"):
                ui.label("Vorhandene Pläne").classes("text-bold")
                with ui.row(wrap=False).classes("gap-0 w-full items-center"):
                    search = (
                        ui.input("Filter")
                        .bind_value(table, "filter")
                        .tooltip("über alle Spalten nach Ausdruck filtern")
                        .props("stack-label clearable square filled dense")
                        .classes("w-full")
                    )
                    with search.add_slot("prepend"):
                        ui.icon("filter_list")
                    ui.button(
                        icon="map", on_click=self._on_load_plan
                    ).bind_enabled_from(
                        table,
                        "selected",
                        backward=lambda selected: (
                            len(selected) > 0 and self.client_is_qgis
                        ),
                    ).tooltip("Selektion in den Layerbaum laden")
                    with (
                        ui.dropdown_button(icon="download")
                        .bind_enabled_from(
                            table,
                            "selected",
                            backward=lambda selected: len(selected) > 0,
                        )
                        .tooltip("Selektion exportieren") as download_button
                    ):
                        ui.item(
                            "GML-Datei",
                            on_click=lambda _: self._on_download_plan(download_button),
                        )
                        ui.item(
                            "ZIP-Archiv",
                            on_click=lambda _: self._on_download_plan(
                                download_button, archive=True
                            ),
                        )
                    ui.button(
                        icon="file_copy", on_click=self._on_copy_plan
                    ).bind_enabled_from(
                        table,
                        "selected",
                        backward=lambda selected: len(selected) > 0,
                    ).tooltip("Kopien von Selektion in Datenbank erzeugen")
                    ui.button(
                        icon="delete", on_click=self._on_delete_plan
                    ).bind_enabled_from(
                        table,
                        "selected",
                        backward=lambda selected: len(selected) > 0,
                    ).tooltip("Selektion aus Datenbank löschen")
                with ui.row(wrap=False).classes("gap-0 w-full items-center"):
                    ui.label("Spalten: ")
                    for column in table.columns:
                        if column["name"] in ["name", "featuretype", "appschema"]:
                            continue
                        ui.switch(
                            column["label"],
                            value=False,
                            on_change=lambda e, column=column: self._on_toggle_column(
                                column, e.value
                            ),
                        )
        return table

    def _on_toggle_column(self, column: dict, visible: bool) -> None:
        column["classes"] = "" if visible else "hidden"
        column["headerClasses"] = "" if visible else "hidden"
        self.plans_table.update()

    async def _on_row_click(self, e: GenericEventArguments):
        """Opens a dialog with action buttons for the row/feature."""
        try:
            row = e.args[1]
            feature_id = row["id"]
        except Exception as e:
            logger.exception("error on loading feature data from row: %s", e)
            return ui.notify("Fehler beim Abruf der Feature-Daten", type="negative")
        self.plans_dialog.clear()
        with self.plans_dialog, ui.card():
            ui.label("Planinformationen").classes("text-bold")
            with ui.grid(columns="auto auto").classes("items-center") as plan_info:
                ui.label("ID")
                ui.label(feature_id)
                ui.label("Name")
                ui.label(row["name"])
                ui.label("Appschema")
                ui.label(row["appschema"])
                ui.label("Aktualisierung")
                ui.label(row["updated"])
                ui.label("Erstellung/Import")
                ui.label(row["created"])
                ui.label("Bereiche")
                ui.label(len(row["bereiche"]))
                ui.label("Anzahl Features")
                spinner = ui.spinner()
            ui.label("Plandokumente").classes("text-bold")
            await DocumentEditor(plan_id=row["id"]).classes("w-full h-full").render()
        self.plans_dialog.open()
        feature_number = await run.io_bound(get_feature_number, feature_id)
        plan_info.remove(spinner)
        ui.label(str(feature_number)).move(plan_info)



from fastapi import HTTPException
from fastapi.responses import FileResponse
from nicegui import run

from xmas_app.settings import get_settings
from xmas_app.util.db import get_doc_by_name


async def get_document_from_db(plan_id: str, filename: str) -> FileResponse:
    """Retrieves a document from the database.

    Supports GET, HEAD as well as range requests, e.g. from GDAL regarding COG.
    """
    file_dir = get_settings().static_file_dir / plan_id

    file_dir.mkdir(exist_ok=True)

    file = file_dir / filename

    if not file.exists():
        doc = await run.io_bound(get_doc_by_name, plan_id, filename)
        if doc:
            file.write_bytes(doc.filedata)
        else:
            raise HTTPException(404)

    return FileResponse(
        path=file,
    )



import json
import logging

from nicegui import app, run, ui
from xplan_tools.model.base import Appschema
from xplan_tools.model.result.result_report import (
    ErrorIssue,
    TransformationValidationError,
    WarningIssue,
)
from xplan_tools.transform import transformer_factory
from xplan_tools.util import MigrationPath

from xmas_app.components.buttons import FeatureInteractionButton
from xmas_app.settings import get_settings

logger = logging.getLogger("xmas_app")


class PlanMigration(ui.column):
    def __init__(self, plan_id: str) -> None:
        super().__init__()
        self.plan_id = plan_id
        self.classes("w-full h-full")
        self.repo = get_settings().repo

    async def render(self) -> None:
        with self:
            self.action_dialog = ui.dialog().on("hide", lambda e: e.sender.clear())
            self.version_select = ui.select([], label="Zielversion auswählen").classes(
                "w-full"
            )
            with self.version_select.add_slot("after"):
                self.start_button = ui.button(
                    icon="play_circle",
                    on_click=lambda: self._on_migrate(self.version_select.value),
                ).bind_enabled_from(self.version_select, "value")
            ui.separator()
            self.spinner_box = ui.row().classes("w-full justify-center items-center")
            self.issue_list = ui.list()
        await self._get_select_options()

    async def _get_select_options(self) -> None:
        plan = await run.io_bound(self.repo.get, self.plan_id)
        plan_appschema = plan.appschema()
        available_versions = {
            appschema.version: f"{plan_appschema.prefix.upper()} {appschema.version}"
            for appschema in Appschema.supported_appschemas()
            if appschema.prefix == plan_appschema.prefix
            and appschema.full_version > plan_appschema.full_version
        }
        self.version_select.set_options(available_versions, value=None)

    async def _on_migrate(self, version: str) -> None:
        with self.spinner_box:
            ui.spinner(size="xl")
        self.start_button.disable()
        self.issue_list.clear()
        try:
            collection = await run.io_bound(self.repo.get_plan_by_id, self.plan_id)
            feature_ids = set(collection.features.keys())
            path = MigrationPath(collection.version, version).path
            if not path:
                ui.notify("Keinen Migrationspfad gefunden", type="warning")
                return
            if len(path) > 1:
                ui.notify(
                    f"Führe mehrschrittige Versionmigration {' - '.join(path)} aus.",
                )
            result = None
            reports = {}
            for step in path:
                try:
                    result = transformer_factory(collection, step).transform()
                    collection = result.output
                    if step == "6.1":
                        collection = transformer_factory(
                            collection, "6.1"
                        ).set_associations()
                    reports[step] = result.warnings
                except TransformationValidationError as e:
                    logger.exception("error during version migration: %r", e)
                    ui.notify(
                        "Bei der Migration sind Validierungsfehler aufgetreten",
                        type="negative",
                    )
                    return self._show_report({step: e.result.errors})
            if not result:
                raise ValueError("no version path found")
            try:
                if deleted := feature_ids.difference(result.output.features.keys()):
                    logger.info(
                        "the following features are deleted after migration to version %s: %s",
                        version,
                        deleted,
                    )
                    for d in deleted:
                        await run.io_bound(self.repo.delete, d)
                await run.io_bound(self.repo.update_all, result.output)
            except Exception as e:
                logger.exception(
                    "error while saving migrated data for plan %r: %s", self.plan_id, e
                )
                ui.notify("Fehler beim Speichern der migrierten Daten", type="negative")
                return
            msg = f"Migration nach {version} erfolgreich"
            if not result.has_warnings:
                ui.notify(msg, type="positive")
            else:
                ui.notify(f"{msg}, Warnungen vorhanden", type="warning")
            # notify plan tree to update nodes
            for client in app.clients("/plan-tools/{id}"):
                client.run_javascript("emitEvent('featureAdded')")
            # notify plan manager to update table
            for client in app.clients("/plans"):
                client.run_javascript("emitEvent('planCreated')")
            logger.info("Sending migrated_plan data to QWebChannel handler")
            data = {"id": self.plan_id, "version": version}
            payload_json = json.dumps(data)
            ui.run_javascript(
                f"""
                    (window.ensureXmasHandler ? window.ensureXmasHandler() : Promise.reject('QWebChannel unavailable'))
                        .then(handler => handler.migrated_plan({payload_json}))
                        .catch(err => console.error('migrated_plan failed', err));
                """
            )
            self._show_report(reports)
        except Exception as e:
            logger.exception(
                "version migration for plan %r failed: %s", self.plan_id, e
            )
            ui.notify("Versionmigration fehlgeschlagen", type="negative")
        finally:
            self.spinner_box.clear()
            await self._get_select_options()
            self.start_button.enable()

    def _show_report(
        self, reports: dict[str, list[WarningIssue] | list[ErrorIssue]]
    ) -> None:
        def copy_issues_to_clipboard():
            ui.clipboard.write(
                "\n".join(
                    issue.message for report in reports.values() for issue in report
                )
            )

        with self.issue_list:
            ui.button(icon="copy_all", on_click=copy_issues_to_clipboard).classes(
                "float-right"
            ).set_visibility(
                False
            )  # disable for now until copying is enabled on Qt side
            for step, issues in reports.items():
                ui.item_label(f"Version {step}").props("header").classes("text-bold")
                if not issues:
                    with ui.item():
                        with ui.item_section().props("avatar"):
                            ui.icon("check_circle", color="positive")
                        with ui.item_section():
                            ui.item_label("keine Warnungen oder Fehler vorhanden")
                else:
                    for issue in issues:
                        with ui.item() as issue_item:
                            if issue.location:
                                issue_item.on_click(
                                    lambda loc=issue.location: self._show_action_menu(
                                        loc
                                    )
                                )
                            with ui.item_section().props("avatar"):
                                if isinstance(issue, WarningIssue):
                                    ui.icon("warning", color="warning")
                                else:
                                    ui.icon("error", color="negative")
                            with ui.item_section():
                                ui.item_label(issue.message)

    async def _show_action_menu(self, feature_id: str):
        feature = await run.io_bound(self.repo.get, feature_id)
        with self.action_dialog, ui.card():
            FeatureInteractionButton(
                "highlight_feature",
                feature_id,
                geom=feature.get_geom_wkt() is not None,
            )
            FeatureInteractionButton("select_feature", feature_id)
            FeatureInteractionButton("show_attribute_form", feature_id)
        self.action_dialog.open()



import asyncio
import logging
import shutil
from contextlib import asynccontextmanager
from importlib import metadata
from pathlib import Path
from tempfile import gettempdir
from uuid import UUID

import pydantic
import pydantic_core
from fastapi import APIRouter, Depends, HTTPException, Request, status
from fastapi.responses import FileResponse
from nicegui import app, ui
from starlette.applications import Starlette

from xmas_app.deps.version_guard import enforce_plugin_version
from xmas_app.form import ModelForm
from xmas_app.models.crud import InsertPayload, UpdatePayload
from xmas_app.models.split import (
    ErrorDetail,
    ErrorResponse,
    SplitPayload,
    SplitSuccess,
)
from xmas_app.pages import PlanManagerPage
from xmas_app.pages.plantools import PlanToolsPage
from xmas_app.services import crud
from xmas_app.services.document import get_document_from_db
from xmas_app.services.split import PlanSplitService, SplitValidationError
from xmas_app.settings import get_settings, initialize_db
from xmas_app.util import get_model_for_select

__version__ = metadata.version("xmas_app")


def _resolve_log_dir() -> Path:
    if get_settings().app_mode == "prod":
        return Path(gettempdir()) / "xmas_log"
    else:
        # dev: local repo logs
        return Path(__file__).parent.parent


log_dir = _resolve_log_dir()
log_dir.mkdir(exist_ok=True)
log_file = log_dir / "xmas_app.log"

logger = logging.getLogger("xmas_app")
logger.propagate = False

logger.handlers.clear()
logger.setLevel(logging.DEBUG if get_settings().debug else logging.INFO)
fh = logging.FileHandler(log_file, mode="w", encoding="utf-8")
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
fh.setFormatter(formatter)
logger.addHandler(fh)
logger.debug(f"Writing logs to {log_file}")

ui.element.default_props("dense")

router = APIRouter()

app.on_startup(initialize_db)
app.on_shutdown(lambda: shutil.rmtree(get_settings().static_file_dir))


@ui.page(
    "/plan-tools/{id}",
    reconnect_timeout=5,
    dependencies=[Depends(enforce_plugin_version)],
)
async def plan_tools(
    request: Request,
    id: UUID,
):
    """Render the PlanTools page.

    Args:
        request: Incoming FastAPI HTTP request to check the user-agent header.
        id: The ID of the plan.
    """
    await PlanToolsPage(request=request, plan_id=str(id)).render()


@ui.page("/plans", dependencies=[Depends(enforce_plugin_version)])
async def plans(
    request: Request,
    appschema: str = get_settings().appschema,
    version: str = get_settings().appschema_version,
):
    """Render the PlanManager page.

    Args:
        request: Incoming FastAPI HTTP request to check the user-agent header.
        appschema (str): Application schema name (default from settings).
        version (str): Schema version (default from settings).
    """
    await PlanManagerPage(
        request=request, appschema=appschema, version=version
    ).render()


@ui.page("/feature/{id}", dependencies=[Depends(enforce_plugin_version)])
async def feature(
    request: Request,
    id: UUID,
    planId: UUID | None = None,
    parentId: UUID | None = None,
    featureType: str | None = None,
    featuretypeRegex: str | None = None,
    appschema: str = get_settings().appschema,
    version: str = get_settings().appschema_version,
):
    """Render and manage a single feature view for the XPlan-GUI application.

    Args:
        request (Request): The incoming FastAPI HTTP request object.
        id (str): UUID or identifier of the feature to display/edit.
        planId (str | None): Optional ID of the associated plan.
        parentId (str | None): Optional ID of the parent feature.
        featureType (str | None): Optional fixed feature type to display.
        featuretypeRegex (str | None): Optional regex to filter allowed feature types.
        appschema (str): The application schema name (defaults from settings).
        version (str): Schema version string (defaults from settings).
    """
    id = str(id)

    async def get_qgis_feature():
        return await ui.run_javascript(
            """return await new Promise((resolve, reject) => {
                    new QWebChannel(qt.webChannelTransport, function (channel) {
                        channel.objects.handler.transfer_feature().then(data => data ? resolve(data) : reject())
                    })
                });""",
            timeout=3,
        )

    async def add_form(
        feature_type: str,
        feature: dict,
    ) -> None:
        content.set_visibility(False)
        spinner.set_visibility(True)
        with content:
            await asyncio.sleep(0.1)  # let spinner transfer state to client
            await app.storage.client["form"].render_form(feature_type, feature)
            if get_settings().debug:
                with ui.row().classes("w-full"):
                    obj_log = ui.log()
                    ui.button(
                        "show model",
                        on_click=lambda: (
                            obj_log.clear(),
                            obj_log.push(
                                model.model_dump_json(indent=2)
                                if (model := app.storage.client["form"].model_instance)
                                else "No Model",
                            ),
                        ),
                    )
                    ui.button(
                        "show submodels",
                        on_click=lambda: (
                            obj_log.clear(),
                            obj_log.push(
                                "\n".join(
                                    model.model_dump_json()
                                    for model in app.storage.client[
                                        "form"
                                    ].sub_models.values()
                                )
                            ),
                        ),
                    )
        spinner.delete()
        content.set_visibility(True)

    def handle_save():
        form: ModelForm = app.storage.client["form"]
        feature = form.model_instance
        if feature:
            feature.id = id
            get_settings().repo.update(feature)
            ui.navigate.to(f"/?saved_feature_uuid={feature.id}")

    def handle_delete():
        form: ModelForm = app.storage.client["form"]
        feature = form.model_instance
        if feature:
            get_settings().repo.delete(id)
            ui.navigate.to("/")

    def on_edit_mode_changed(e):
        if form := app.storage.client["form"]:
            form.editable = e.args

    ui.colors(primary="rgb(157 157 156)", secondary="rgb(57 92 127)")
    app.storage.client["form"] = None
    app.storage.client["plan_id"] = str(planId) if planId else None
    app.storage.client["parent_id"] = str(parentId) if parentId else None
    app.storage.client["user_agent"] = request.headers.get("user-agent", None)

    qgis = app.storage.client["user_agent"].startswith(get_settings().qgis_plugin_name)

    if qgis:
        ui.add_head_html('<script src="qrc:///qtwebchannel/qwebchannel.js"></script>')

    try:
        model = get_settings().repo.get(id)
        feature_type = model.get_name()
        feature = model.model_dump(
            mode="json", context={"datatype": True, "file_uri": True}
        )
    except ValueError:
        feature = {"id": id}
        feature_type = featureType

    ui.on("editModeChanged", on_edit_mode_changed)

    header = ui.header()  # .classes("bg-white text-black")

    if not qgis:
        with header:
            with ui.column(align_items="center").bind_visibility_from(
                app.storage.client,
                "form",
                backward=lambda form: getattr(form, "rendered", False),
            ):
                with ui.row(align_items="center"):
                    ui.button(
                        text="Speichern",
                        icon="save",
                        on_click=handle_save,
                    ).bind_enabled_from(app.storage.client, "form").props(
                        "unelevated rounded no-caps"
                    )
                    ui.button(
                        text="Löschen",
                        icon="delete",
                        on_click=handle_delete,
                    ).bind_enabled_from(app.storage.client, "form").props(
                        "unelevated rounded no-caps"
                    )
                    ui.button(
                        text="Abbrechen",
                        icon="cancel",
                        on_click=lambda: ui.navigate.to("/"),
                    ).props("unelevated rounded no-caps")
    spinner = ui.spinner(size="xl").classes("absolute-center")
    with ui.grid(columns=2 if get_settings().debug else 1).classes(
        "justify-center items-start"
    ) as content:
        with ui.column(align_items="center").classes(
            "justify-center"
        ):  # .classes("absolute-center"):
            with ui.row(align_items="center").classes("col-grow") as sel:
                ui.label("Objektart auswählen").classes("text-h6")
                model_select = (
                    ui.select(
                        options=[],
                        value=None,
                        label="Objektarten",
                        on_change=lambda x: add_form(x.value, feature),
                        with_input=True,
                    )
                    .props("square filled options-dense")
                    .style("width: 500px;")
                )
                # .classes("absolute-center")
            if feature_type:
                sel.set_visibility(False)
            else:
                spinner.set_visibility(False)
        app.storage.client["form"] = ModelForm(appschema, version, content, header)
        await ui.context.client.connected(
            timeout=10
        )  # see https://nicegui.io/documentation/page#wait_for_client_connection
        if qgis:
            qgis_feature = await get_qgis_feature()
            if isinstance(properties := qgis_feature.get("properties"), dict):
                feature = feature | properties
            if isinstance(geometry := qgis_feature.get("geometry"), dict):
                feature["geometry"] = geometry
        if feature_type:
            await add_form(feature_type, feature)
        else:
            await get_model_for_select(
                model_select,
                qgis_feature.get("geometry", {}).get("wkt", None) if qgis else None,
                featuretypeRegex,
                appschema,
                version,
            )


@app.get("/health_check")
def health_check() -> dict:
    """Health check endpoint to verify that the service is running.

    Returns:
        dict: A simple dictionary with the app's name and version.
    """
    return {"name": "XMAS-App", "version": __version__}


@app.api_route("/plan/{plan_id}/documents/{filename}", methods=["GET", "HEAD"])
async def get_document_data(plan_id: UUID, filename: str) -> FileResponse:
    """Returns the data of a document."""
    response = await get_document_from_db(str(plan_id), filename)
    return response


@app.get("/feature/{id}/properties", dependencies=[Depends(enforce_plugin_version)])
async def get_feature_data(id: UUID) -> dict:
    """Returns the data of a feature."""
    feature = get_settings().repo.get(str(id))
    excludes = {"id"}
    if geom_field := feature.get_geom_field():
        excludes.add(geom_field)
    return feature.model_dump(mode="json", exclude=excludes, exclude_unset=True)


@app.post(
    "/insert-features", status_code=201, dependencies=[Depends(enforce_plugin_version)]
)
async def insert_features(payload: InsertPayload, planId: UUID):
    """Insert a number of features."""
    await crud.create(payload, str(planId))


@app.post(
    "/update-features", status_code=204, dependencies=[Depends(enforce_plugin_version)]
)
async def update_features(payload: UpdatePayload):
    """Update a number of features."""
    await crud.update(payload)


@app.post(
    "/split-tool",
    response_model=SplitSuccess,
    status_code=status.HTTP_201_CREATED,  # semantically 'created'
    responses={
        400: {"model": ErrorResponse, "description": "Bad Request"},
        422: {"model": ErrorResponse, "description": "Validation Error"},
        500: {"model": ErrorResponse, "description": "Server Error"},
    },
    dependencies=[Depends(enforce_plugin_version)],
)
async def receive_split_plans(payload: SplitPayload) -> SplitSuccess:
    logger.info("Split plans endpoint reached.")

    API_VERSION = "split-import-v1-min"

    # Lightweight contract check up front
    if payload.schema_version != API_VERSION:
        logger.error("Payload api schema version incorrect.")
        raise HTTPException(
            status_code=400,
            detail=ErrorDetail(
                code="invalid_schema_version", message="Invalid schema_version"
            ).model_dump(),
        )

    # Early existence check to short-circuit before the service
    try:
        source_plan = get_settings().repo.get_plan_by_id(payload.src_plan_id)
        if source_plan is None:
            logger.error("Couldn't load plan from database with provided id.")
            raise HTTPException(
                status_code=400,
                detail=ErrorDetail(
                    code="plan_not_found", message="Plan not found"
                ).model_dump(),
            )
    except (pydantic.ValidationError, pydantic_core.ValidationError) as e:
        logger.error(f"Plan validation failed: {e}")
        raise HTTPException(
            status_code=422,
            detail=ErrorDetail(code="validation_error", message=str(e)).model_dump(),
        )

    service = PlanSplitService(get_settings().repo)

    try:
        return service.apply_split_import(payload, source_plan)
    except SplitValidationError as e:
        logger.warning("Split validation failed with %d violations", len(e.violations))
        raise HTTPException(
            status_code=400,
            detail=ErrorDetail(
                code="validation_failed",
                message=e.message,
                violations=[
                    {
                        "code": v.code,
                        "featuretype": v.featuretype,
                        "old_id": v.old_id,
                        "hint": v.hint,
                    }
                    for v in e.violations
                ],
            ).model_dump(),
        )
    except ValueError as e:
        logger.error("Domain error during split: %s", e)
        raise HTTPException(
            status_code=400,
            detail=ErrorDetail(code="domain_error", message=str(e)).model_dump(),
        )
    except Exception as e:
        logger.exception(
            "Unhandled split processing error. "
            "Likely dangling association reference. "
            "Check for edges whose dst_old not in id_map."
        )
        raise HTTPException(
            status_code=500,
            detail=ErrorDetail(
                code="server_error", message=f"Processing error: {e}"
            ).model_dump(),
        ) from e


@asynccontextmanager
async def lifespan(starlette_app):
    msg = f"XMAS-App is running on port {get_settings().app_port} in {get_settings().app_mode} mode."
    logger.info(msg)
    print(msg)
    yield
    logger.info("XMAS-App is shutting down.")
    print("XMAS-App is shutting down.")


starlette_app = Starlette(
    debug=get_settings().debug,
    routes=None,
    lifespan=lifespan,
)
ui.run_with(starlette_app)


def create_app() -> Starlette:
    """Factory for the plugin: return the ASGI app, without asyncio or uvicorn."""

    # os.environ["PYGEOAPI_CONFIG"] = str(Path(__file__).parent / "pygeoapi" / "config.yaml")
    # os.environ["PYGEOAPI_OPENAPI"] = str(
    #     Path(__file__).parent / "pygeoapi" / "openapi.yaml"
    # )
    # from pygeoapi.starlette_app import APP as pygeoapi_app

    # starlette.mount("/oapi", pygeoapi_app)

    return starlette_app


def run_server():
    import uvicorn

    print("Starting server in stand alone mode.")
    logger.info("Starting server in stand alone mode.")

    try:
        uvicorn.run(
            "xmas_app.main:starlette_app",
            host="0.0.0.0",
            port=get_settings().app_port,
            reload=get_settings().app_mode == "dev",
            log_config=None,  # avoids isatty() issue
        )
    except Exception as e:
        msg = f"Failed to start server: {e}"
        print(msg)
        logger.error(msg)


if __name__ == "__main__":
    run_server()



"""Adds a trigger to notify QGIS on changes."""

from alembic import op
from xplan_tools.settings import get_settings

revision = "3715fd158e35"
down_revision = None
branch_labels = None
depends_on = None


schema = get_settings().db_schema or "public"


def upgrade():
    """Upgrade schema."""
    op.execute(f"""
        DROP TRIGGER IF EXISTS trg_notify_qgis ON {schema}.coretable;
        DROP FUNCTION IF EXISTS {schema}.fn_notify_qgis();

        CREATE OR REPLACE FUNCTION {schema}.xmas_notify_qgis() RETURNS trigger AS $$
        BEGIN
            PERFORM pg_notify('qgis', concat('xmas_plugin/refresh_', NEW.geometry_type));
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;

        CREATE TRIGGER xmas_notify_qgis
        AFTER INSERT OR UPDATE ON {schema}.coretable
        FOR EACH ROW
        EXECUTE FUNCTION {schema}.xmas_notify_qgis();
    """)


def downgrade():
    """Downgrade schema."""
    op.execute(f"""
        DROP TRIGGER IF EXISTS xmas_notify_qgis ON {schema}.coretable;
        DROP FUNCTION IF EXISTS {schema}.xmas_notify_qgis();
    """)



from fastapi import Request
from nicegui import binding, ui

from xmas_app.settings import get_settings


class BasePage:
    """Base class providing some defaults."""

    client_is_qgis = binding.BindableProperty()

    def __init__(self, request: Request) -> None:
        """Initializes the BasePage.

        Args:
            request: a `fastapi.Request` for e.g. header extraction
        """
        ui.colors(primary="rgb(157 157 156)", secondary="rgb(57 92 127)")
        ui.button.default_props("flat square no-caps")
        ui.dropdown_button.default_props("flat square no-caps")
        ui.select.default_props("square filled dense")
        request.headers.get("user-agent", None)
        self.client_is_qgis = request.headers.get("user-agent", "").startswith(
            get_settings().qgis_plugin_name
        )
        if self.client_is_qgis:
            ui.add_head_html(
                '<script src="qrc:///qtwebchannel/qwebchannel.js"></script>'
            )



from alembic import context

from xmas_app.models.orm import Base

config = context.config

target_metadata = Base.metadata


def run_migrations_offline():
    context.configure(
        url=None,
        target_metadata=target_metadata,
        literal_binds=True,
        compare_type=True,
        version_table="xmas_app_revision",
    )
    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    connection = config.attributes.get("connection")
    if connection is not None:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table="xmas_app_revision",
        )
        with context.begin_transaction():
            context.run_migrations()
    else:
        from xmas_app.settings import get_settings

        repo = get_settings().repo
        config.attributes["schema"] = repo.schema or "public"
        with repo._engine.connect() as connection:
            context.configure(
                connection=connection,
                target_metadata=target_metadata,
                version_table="xmas_app_revision",
            )
            with context.begin_transaction():
                context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import logging

from nicegui import app, events, run, ui
from pydantic import AnyUrl, ValidationError

from xmas_app.settings import get_settings
from xmas_app.util.db import (
    delete_documents,
    get_doc_by_name,
    get_plan_documents,
    save_documents,
)

logger = logging.getLogger("xmas_app")


class DocumentSelect(ui.select):
    def __init__(self, bind_obj: object, key: str, label: str, validation: dict):
        initial_options = []
        if value := getattr(bind_obj, key, None):
            value = self._ensure_file_uri(value)
            initial_options.append(value)
        super().__init__(
            initial_options,
            label=label,
            value=value,
            new_value_mode="add-unique",
            validation=validation,
        )
        self.bind_value(
            bind_obj,
            key,
            backward=self._ensure_file_uri,
            forward=self._ensure_file_uri,
        )
        self.on("popup-show", self._get_select_options)
        self.on(
            "new-value",
            lambda e: self.set_options(
                self.options + [self._ensure_file_uri(e.args[0])]
            ),
        )

    async def _get_select_options(self):
        options = set(self.options)
        docs = await run.io_bound(get_plan_documents, app.storage.client["plan_id"])
        for doc in docs:
            options.add(doc)
        self.set_options(list(options))

    def _ensure_file_uri(self, value: str | None) -> str | None:
        if not value:
            return None
        try:
            AnyUrl(value)
        except ValidationError:
            return f"file:///{value}"
        else:
            return value


class DocumentEditor(ui.column):
    def __init__(
        self,
        *,
        plan_id: str,
    ):
        super().__init__()
        self.plan_id = plan_id

    async def render(self):
        with self:
            self.docs_table = self._create_table()
            ui.upload(
                multiple=True,
                max_total_size=50000000,
                label="Dokumente importieren",
                on_multi_upload=self._on_upload,
            ).props("flat square bordered").classes("w-full")
        await self._update_table()

    def _create_table(self):
        columns = [
            {
                "name": "filename",
                "label": "Dateiname",
                "field": "filename",
                "align": "left",
                "sortable": True,
            },
            {
                "name": "filetype",
                "label": "Dateityp",
                "field": "filetype",
                "align": "left",
                "sortable": True,
            },
        ]
        table = (
            ui.table(
                rows=[],
                columns=columns,
                row_key="pk",
                selection="multiple",
            )
            .props(
                """
                no-data-label='keine Dokumente vorhanden'
                no-results-label='keine passenden Ergebnisse gefunden'
                flat
                square
                dense
                virtual-scroll
                bordered
            """
            )
            .classes("w-full h-full")
        )
        table.on("row-click", self._on_row_click)
        with table.add_slot("top"):
            with ui.row(wrap=False).classes("gap-0 w-full"):
                search = (
                    ui.input("Filter")
                    .bind_value(table, "filter")
                    .tooltip("über alle Spalten nach Ausdruck filtern")
                    .props("stack-label clearable square filled dense")
                    .classes("w-full")
                )
                with search.add_slot("prepend"):
                    ui.icon("filter_list")
                ui.button(
                    icon="delete", on_click=self._on_delete_docs
                ).bind_enabled_from(
                    table,
                    "selected",
                    backward=lambda selected: len(selected) > 0,
                ).tooltip("Selektion aus Datenbank löschen")
        return table

    async def _on_upload(self, event: events.MultiUploadEventArguments):
        self.docs_table.props("loading")
        try:
            data = []
            for index, content in enumerate(event.contents):
                data.append(
                    {
                        "id": self.plan_id,
                        "filename": event.names[index],
                        "filetype": event.types[index],
                        "filedata": content.read(),
                    }
                )
            save_documents(data)
            await self._update_table()
            self._notify_plan_table()
        except Exception as e:
            logger.exception("error uploading documents: %r", e)
            ui.notify("Fehler beim Speichern der Dokumente", type="negative")

    async def _update_table(self):
        try:
            self.docs_table.props("loading")
            docs = await run.io_bound(get_plan_documents, self.plan_id, True)
            self.docs_table.update_rows(docs, clear_selection=False)
            self._update_row_select_label()
        finally:
            self.docs_table.props(remove="loading")

    def _notify_plan_table(self):
        """Notify the plan table that documents have changed."""
        for client in app.clients("/plans"):
            client.run_javascript("emitEvent('planCreated')")

    async def _on_delete_docs(self):
        try:
            for doc in self.docs_table.selected:
                (
                    get_settings().static_file_dir / self.plan_id / doc["filename"]
                ).unlink(missing_ok=True)
            delete_documents([doc["pk"] for doc in self.docs_table.selected])
            await self._update_table()
            self._notify_plan_table()
        except Exception as e:
            logger.exception("error deleting documents: %r", e)
            ui.notify("Fehler beim Löschen der Dokumente", type="negative")

    def _on_download_file(self):
        doc = self.docs_table.selected[0]
        ui.download.content(
            get_doc_by_name(self.plan_id, doc["filename"]).filedata,
            filename=doc["filename"],
            media_type=doc["filetype"],
        )

    def _update_row_select_label(self):
        doc_number = len(self.docs_table.rows)
        self.docs_table.props(
            f":selected-rows-label='(numberOfRows) => `${{ numberOfRows }} von {doc_number} Dokumenten selektiert`'"
        )

    async def _on_row_click(self, event: events.GenericEventArguments):
        try:
            row = event.args[1]
            doc_name = row["filename"]
        except Exception as e:
            logger.exception("error on loading document data from row: %s", e)
            return ui.notify("Fehler beim Abruf der Dokument-Daten", type="negative")
        ui.navigate.to(f"/plan/{self.plan_id}/documents/{doc_name}")
